#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unittesting the Native SEVIRI reader."""

from __future__ import annotations

import datetime as dt
import os
import unittest
import warnings
import zipfile
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from pytest_lazy_fixtures import lf

from satpy.readers.core.eum import recarray2dict, time_cds_short
from satpy.readers.seviri_l1b_native import (
    ASCII_STARTSWITH,
    ImageBoundaries,
    NativeMSGFileHandler,
    Padder,
    get_available_channels,
    has_archive_header,
)
from satpy.readers.seviri_l1b_native_hdr import Msg15NativeHeaderRecord
from satpy.scene import Scene
from satpy.tests.reader_tests.test_seviri_base import ORBIT_POLYNOMIALS, ORBIT_POLYNOMIALS_INVALID
from satpy.tests.reader_tests.test_seviri_l1b_calibration import TestFileHandlerCalibrationBase
from satpy.tests.utils import assert_attrs_equal, make_dataid

CHANNEL_INDEX_LIST = ["VIS006", "VIS008", "IR_016", "IR_039",
                      "WV_062", "WV_073", "IR_087", "IR_097",
                      "IR_108", "IR_120", "IR_134", "HRV"]
AVAILABLE_CHANNELS = {}
for item in CHANNEL_INDEX_LIST:
    AVAILABLE_CHANNELS[item] = True

SEC15HDR = "15_SECONDARY_PRODUCT_HEADER"
IDS = "SelectedBandIDs"

TEST1_HEADER_CHNLIST: dict[str, dict[str, dict]] = {SEC15HDR: {IDS: {}}}
TEST1_HEADER_CHNLIST[SEC15HDR][IDS]["Value"] = "XX--XX--XX--"

TEST2_HEADER_CHNLIST: dict[str, dict[str, dict]] = {SEC15HDR: {IDS: {}}}
TEST2_HEADER_CHNLIST[SEC15HDR][IDS]["Value"] = "XX-XXXX----X"

TEST3_HEADER_CHNLIST: dict[str, dict[str, dict]] = {SEC15HDR: {IDS: {}}}
TEST3_HEADER_CHNLIST[SEC15HDR][IDS]["Value"] = "XXXXXXXXXXXX"

TEST_AREA_EXTENT_EARTHMODEL1_VISIR_FULLDISK = {
    "earth_model": 1,
    "dataset_id": make_dataid(name="VIS006", resolution=3000),
    "is_full_disk": True,
    "is_rapid_scan": 0,
    "fill_disk": False,
    "expected_area_def": {
        "Area ID": "msg_seviri_fes_3km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "0", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 3712,
        "Number of rows": 3712,
        "Area extent": (5568748.2758, 5568748.2758, -5568748.2758, -5568748.2758)
    }
}

TEST_AREA_EXTENT_EARTHMODEL1_VISIR_RAPIDSCAN = {
    "earth_model": 1,
    "dataset_id": make_dataid(name="VIS006", resolution=3000),
    "is_full_disk": False,
    "is_rapid_scan": 1,
    "fill_disk": False,
    "expected_area_def": {
        "Area ID": "msg_seviri_rss_3km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "9.5", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 3712,
        "Number of rows": 1392,
        "Area extent": (5568748.275756836, 5568748.275756836, -5568748.275756836, 1392187.068939209)
    }
}

TEST_AREA_EXTENT_EARTHMODEL1_VISIR_RAPIDSCAN_FILL = {
    "earth_model": 1,
    "dataset_id": make_dataid(name="VIS006", resolution=3000),
    "is_full_disk": False,
    "is_rapid_scan": 1,
    "fill_disk": True,
    "expected_area_def": {
        "Area ID": "msg_seviri_rss_3km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "9.5", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 3712,
        "Number of rows": 3712,
        "Area extent": (5568748.2758, 5568748.2758, -5568748.2758, -5568748.2758)
    }
}

TEST_AREA_EXTENT_EARTHMODEL1_VISIR_ROI = {
    "earth_model": 1,
    "dataset_id": make_dataid(name="VIS006", resolution=3000),
    "is_full_disk": False,
    "is_rapid_scan": 0,
    "fill_disk": False,
    "expected_area_def": {
        "Area ID": "msg_seviri_fes_3km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "0", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 2516,
        "Number of rows": 1829,
        "Area extent": (5337717.232, 5154692.6389, -2211297.1332, -333044.7514)

    }
}

TEST_AREA_EXTENT_EARTHMODEL1_VISIR_ROI_FILL = {
    "earth_model": 1,
    "dataset_id": make_dataid(name="VIS006", resolution=3000),
    "is_full_disk": False,
    "is_rapid_scan": 0,
    "fill_disk": True,
    "expected_area_def": {
        "Area ID": "msg_seviri_fes_3km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "0", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 3712,
        "Number of rows": 3712,
        "Area extent": (5568748.2758, 5568748.2758, -5568748.2758, -5568748.2758)

    }
}

TEST_AREA_EXTENT_EARTHMODEL1_HRV_FULLDISK = {
    "earth_model": 1,
    "dataset_id": make_dataid(name="HRV", resolution=1000),
    "is_full_disk": True,
    "is_rapid_scan": 0,
    "fill_disk": False,
    "expected_area_def": {
        "Area ID": "msg_seviri_fes_1km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "0", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 5568,
        "Number of rows": 11136,
        "Area extent 0": (5567747.920155525, 2625352.665781975, -1000.1343488693237, -5567747.920155525),
        "Area extent 1": (3602483.924627304, 5569748.188853264, -1966264.1298770905, 2625352.665781975)
    }
}

TEST_AREA_EXTENT_EARTHMODEL1_HRV_FULLDISK_FILL = {
    "earth_model": 1,
    "dataset_id": make_dataid(name="HRV", resolution=1000),
    "is_full_disk": True,
    "is_rapid_scan": 0,
    "fill_disk": True,
    "expected_area_def": {
        "Area ID": "msg_seviri_fes_1km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "0", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 11136,
        "Number of rows": 11136,
        "Area extent": (5567747.920155525, 5569748.188853264, -5569748.188853264, -5567747.920155525)
    }
}

TEST_AREA_EXTENT_EARTHMODEL1_HRV_RAPIDSCAN = {
    "earth_model": 1,
    "dataset_id": make_dataid(name="HRV", resolution=1000),
    "is_full_disk": False,
    "is_rapid_scan": 1,
    "fill_disk": False,
    "expected_area_def": {
        "Area ID": "msg_seviri_rss_1km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "9.5", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 5568,
        "Number of rows": 8192,
        "Area extent": (5567747.920155525, 2625352.665781975, -1000.1343488693237, -5567747.920155525)
    }
}

TEST_AREA_EXTENT_EARTHMODEL1_HRV_RAPIDSCAN_FILL = {
    "earth_model": 1,
    "dataset_id": make_dataid(name="HRV", resolution=1000),
    "is_full_disk": False,
    "is_rapid_scan": 1,
    "fill_disk": True,
    "expected_area_def": {
        "Area ID": "msg_seviri_rss_1km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "9.5", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 11136,
        "Number of rows": 11136,
        "Area extent": (5567747.920155525, 5569748.188853264, -5569748.188853264, -5567747.920155525)
    }
}

TEST_AREA_EXTENT_EARTHMODEL1_HRV_ROI = {
    "earth_model": 1,
    "dataset_id": make_dataid(name="HRV", resolution=1000),
    "is_full_disk": False,
    "is_rapid_scan": 0,
    "fill_disk": False,
    "expected_area_def": {
        "Area ID": "msg_seviri_fes_1km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "0", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 7548,
        "Number of rows": 5487,
        "Area extent": (5336716.885566711, 5155692.568421364, -2212297.179698944, -332044.6038246155)
    }
}

TEST_AREA_EXTENT_EARTHMODEL1_HRV_ROI_FILL = {
    "earth_model": 1,
    "dataset_id": make_dataid(name="HRV", resolution=1000),
    "is_full_disk": False,
    "is_rapid_scan": 0,
    "fill_disk": True,
    "expected_area_def": {
        "Area ID": "msg_seviri_fes_1km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "0", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 11136,
        "Number of rows": 11136,
        "Area extent": (5567747.920155525, 5569748.188853264, -5569748.188853264, -5567747.920155525)
    }
}

TEST_AREA_EXTENT_EARTHMODEL2_VISIR_FULLDISK = {
    "earth_model": 2,
    "dataset_id": make_dataid(name="VIS006", resolution=3000),
    "is_full_disk": True,
    "is_rapid_scan": 0,
    "fill_disk": False,
    "expected_area_def": {
        "Area ID": "msg_seviri_fes_3km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "0", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 3712,
        "Number of rows": 3712,
        "Area extent": (5567248.0742, 5570248.4773, -5570248.4773, -5567248.0742)
    }
}

TEST_AREA_EXTENT_EARTHMODEL2_HRV_FULLDISK = {
    "earth_model": 2,
    "dataset_id": make_dataid(name="HRV", resolution=1000),
    "is_full_disk": True,
    "is_rapid_scan": 0,
    "fill_disk": False,
    "expected_area_def": {
        "Area ID": "msg_seviri_fes_1km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "0", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 5568,
        "Number of rows": 11136,
        "Area extent 0": (5566247.718632221, 2626852.867305279, -2500.3358721733093, -5566247.718632221),
        "Area extent 1": (3600983.723104, 5571248.390376568, -1967764.3314003944, 2626852.867305279)
    }
}

TEST_AREA_EXTENT_EARTHMODEL2_HRV_FULLDISK_FILL = {
    "earth_model": 2,
    "dataset_id": make_dataid(name="HRV", resolution=1000),
    "is_full_disk": True,
    "is_rapid_scan": 0,
    "fill_disk": True,
    "expected_area_def": {
        "Area ID": "msg_seviri_fes_1km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "0", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 11136,
        "Number of rows": 11136,
        "Area extent": (5566247.718632221, 5571248.390376568, -5571248.390376568, -5566247.718632221)
    }
}

TEST_AREA_EXTENT_EARTHMODEL2_VISIR_RAPIDSCAN = {
    "earth_model": 2,
    "dataset_id": make_dataid(name="VIS006", resolution=3000),
    "is_full_disk": False,
    "is_rapid_scan": 1,
    "fill_disk": False,
    "expected_area_def": {
        "Area ID": "msg_seviri_rss_3km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "9.5", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 3712,
        "Number of rows": 1392,
        "Area extent": (5567248.074173927, 5570248.477339745, -5570248.477339745, 1393687.2705221176)

    }
}

TEST_AREA_EXTENT_EARTHMODEL2_VISIR_RAPIDSCAN_FILL = {
    "earth_model": 2,
    "dataset_id": make_dataid(name="VIS006", resolution=3000),
    "is_full_disk": False,
    "is_rapid_scan": 1,
    "fill_disk": True,
    "expected_area_def": {
        "Area ID": "msg_seviri_rss_3km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "9.5", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 3712,
        "Number of rows": 3712,
        "Area extent": (5567248.0742, 5570248.4773, -5570248.4773, -5567248.0742)

    }
}

TEST_AREA_EXTENT_EARTHMODEL2_HRV_RAPIDSCAN = {
    "earth_model": 2,
    "dataset_id": make_dataid(name="HRV", resolution=1000),
    "is_full_disk": False,
    "is_rapid_scan": 1,
    "fill_disk": False,
    "expected_area_def": {
        "Area ID": "msg_seviri_rss_1km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "0", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 5568,
        "Number of rows": 8192,
        "Area extent": (5566247.718632221, 2626852.867305279, -2500.3358721733093, -5566247.718632221)
    }
}

TEST_AREA_EXTENT_EARTHMODEL2_HRV_RAPIDSCAN_FILL = {
    "earth_model": 2,
    "dataset_id": make_dataid(name="HRV", resolution=1000),
    "is_full_disk": False,
    "is_rapid_scan": 1,
    "fill_disk": True,
    "expected_area_def": {
        "Area ID": "msg_seviri_rss_1km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "0", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 11136,
        "Number of rows": 11136,
        "Area extent": (5566247.718632221, 5571248.390376568, -5571248.390376568, -5566247.718632221)
    }
}

TEST_AREA_EXTENT_EARTHMODEL2_VISIR_ROI = {
    "earth_model": 2,
    "dataset_id": make_dataid(name="VIS006", resolution=3000),
    "is_full_disk": False,
    "is_rapid_scan": 0,
    "fill_disk": False,
    "expected_area_def": {
        "Area ID": "msg_seviri_fes_3km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "0", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 2516,
        "Number of rows": 1829,
        "Area extent": (5336217.0304, 5156192.8405, -2212797.3348, -331544.5498)
    }
}

TEST_AREA_EXTENT_EARTHMODEL2_VISIR_ROI_FILL = {
    "earth_model": 2,
    "dataset_id": make_dataid(name="VIS006", resolution=3000),
    "is_full_disk": False,
    "is_rapid_scan": 0,
    "fill_disk": True,
    "expected_area_def": {
        "Area ID": "msg_seviri_fes_3km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "0", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 3712,
        "Number of rows": 3712,
        "Area extent": (5567248.0742, 5570248.4773, -5570248.4773, -5567248.0742)
    }
}

TEST_AREA_EXTENT_EARTHMODEL2_HRV_ROI = {
    "earth_model": 2,
    "dataset_id": make_dataid(name="HRV", resolution=1000),
    "is_full_disk": False,
    "is_rapid_scan": 0,
    "fill_disk": False,
    "expected_area_def": {
        "Area ID": "msg_seviri_fes_1km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "0", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 7548,
        "Number of rows": 5487,
        "Area extent": (5335216.684043407, 5157192.769944668, -2213797.381222248, -330544.4023013115)
    }
}

TEST_AREA_EXTENT_EARTHMODEL2_HRV_ROI_FILL = {
    "earth_model": 2,
    "dataset_id": make_dataid(name="HRV", resolution=1000),
    "is_full_disk": False,
    "is_rapid_scan": 0,
    "fill_disk": True,
    "expected_area_def": {
        "Area ID": "msg_seviri_fes_1km",
        "Projection": {"a": "6378169000", "b": "6356583800", "h": "35785831",
                       "lon_0": "0", "no_defs": "None", "proj": "geos",
                       "type": "crs", "units": "m", "x_0": "0", "y_0": "0"},
        "Number of columns": 11136,
        "Number of rows": 11136,
        "Area extent": (5566247.718632221, 5571248.390376568, -5571248.390376568, -5566247.718632221)
    }
}

TEST_IS_ROI_FULLDISK = {
    "is_full_disk": True,
    "is_rapid_scan": 0,
    "is_roi": False
}

TEST_IS_ROI_RAPIDSCAN = {
    "is_full_disk": False,
    "is_rapid_scan": 1,
    "is_roi": False
}

TEST_IS_ROI_ROI = {
    "is_full_disk": False,
    "is_rapid_scan": 0,
    "is_roi": True
}

TEST_CALIBRATION_MODE = {
    "earth_model": 1,
    "dataset_id": make_dataid(name="IR_108", calibration="radiance"),
    "is_full_disk": True,
    "is_rapid_scan": 0,
    "calibration": "radiance",
    "CalSlope": [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.95, 0.96, 0.97],
    "CalOffset": [-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0],
    "GSICSCalCoeff": [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.95, 0.96, 0.97],
    "GSICSOffsetCount": [-51.0, -51.0, -51.0, -51.0, -51.0, -51.0, -51.0, -51.0, -51.0, -51.0, -51.0, -51.0]
}

TEST_PADDER_RSS_ROI = {
    "img_bounds": {"south": [2], "north": [4], "east": [2], "west": [3]},
    "is_full_disk": False,
    "dataset_id": make_dataid(name="VIS006"),
    "dataset": xr.DataArray(np.ones((3, 2)), dims=["y", "x"]).astype(np.float32),
    "final_shape": (5, 5),
    "expected_padded_data": xr.DataArray(np.array([[np.nan, np.nan, np.nan, np.nan, np.nan],
                                                   [np.nan, 1.0, 1.0, np.nan, np.nan],
                                                   [np.nan, 1.0, 1.0, np.nan, np.nan],
                                                   [np.nan, 1.0, 1.0, np.nan, np.nan],
                                                   [np.nan, np.nan, np.nan, np.nan, np.nan]]),
                                         dims=["y", "x"]).astype(np.float32)
}

TEST_PADDER_FES_HRV = {
    "img_bounds": {"south": [1, 4], "north": [3, 5], "east": [2, 3], "west": [3, 4]},
    "is_full_disk": True,
    "dataset_id": make_dataid(name="HRV"),
    "dataset": xr.DataArray(np.ones((5, 2)), dims=["y", "x"]).astype(np.float32),
    "final_shape": (5, 5),
    "expected_padded_data": xr.DataArray(np.array([[np.nan, 1.0, 1.0, np.nan, np.nan],
                                                   [np.nan, 1.0, 1.0, np.nan, np.nan],
                                                   [np.nan, 1.0, 1.0, np.nan, np.nan],
                                                   [np.nan, np.nan, 1.0, 1.0, np.nan],
                                                   [np.nan, np.nan, 1.0, 1.0, np.nan]]),
                                         dims=["y", "x"]).astype(np.float32)

}


def create_test_header(earth_model, dataset_id, is_full_disk, is_rapid_scan, good_qual="OK"):
    """Create test header for SEVIRI L1.5 product.

    Header includes mandatory attributes for NativeMSGFileHandler.get_area_extent
    """
    if dataset_id["name"] == "HRV":
        reference_grid = "ReferenceGridHRV"
        column_dir_grid_step = 1.0001343488693237
        line_dir_grid_step = 1.0001343488693237
    else:
        reference_grid = "ReferenceGridVIS_IR"
        column_dir_grid_step = 3.0004031658172607
        line_dir_grid_step = 3.0004031658172607

    if is_full_disk:
        north = 3712
        east = 1
        west = 3712
        south = 1
        n_visir_cols = 3712
        n_visir_lines = 3712
        n_hrv_cols = 11136
        n_hrv_lines = 11136
        ssp_lon = 0
    elif is_rapid_scan:
        north = 3712
        east = 1
        west = 3712
        south = 2321
        n_visir_cols = 3712
        n_visir_lines = 1392
        n_hrv_cols = 11136
        n_hrv_lines = 4176
        ssp_lon = 9.5
    else:
        north = 3574
        east = 78
        west = 2591
        south = 1746
        n_visir_cols = 2516
        n_visir_lines = north - south + 1
        n_hrv_cols = n_visir_cols * 3
        n_hrv_lines = n_visir_lines * 3
        ssp_lon = 0
    header = {
        "15_MAIN_PRODUCT_HEADER": {
            "QQOV": {"Name": "QQOV",
                     "Value": good_qual}
        },
        "15_DATA_HEADER": {
            "ImageDescription": {
                reference_grid: {
                    "ColumnDirGridStep": column_dir_grid_step,
                    "LineDirGridStep": line_dir_grid_step,
                    "GridOrigin": 2,  # south-east corner
                },
                "ProjectionDescription": {
                    "LongitudeOfSSP": ssp_lon
                }
            },
            "GeometricProcessing": {
                "EarthModel": {
                    "TypeOfEarthModel": earth_model,
                    "EquatorialRadius": 6378169.0,
                    "NorthPolarRadius": 6356583.800000001,
                    "SouthPolarRadius": 6356583.800000001,
                }
            },
            "SatelliteStatus": {
                "SatelliteDefinition": {
                    "SatelliteId": 324
                }
            }
        },
        "15_SECONDARY_PRODUCT_HEADER": {
            "NorthLineSelectedRectangle": {"Value": north},
            "EastColumnSelectedRectangle": {"Value": east},
            "WestColumnSelectedRectangle": {"Value": west},
            "SouthLineSelectedRectangle": {"Value": south},
            "SelectedBandIDs": {"Value": "xxxxxxxxxxxx"},
            "NumberColumnsVISIR": {"Value": n_visir_cols},
            "NumberLinesVISIR": {"Value": n_visir_lines},
            "NumberColumnsHRV": {"Value": n_hrv_cols},
            "NumberLinesHRV": {"Value": n_hrv_lines},
        }

    }

    return header


def create_test_trailer(is_rapid_scan):
    """Create test trailer for SEVIRI L1.5 product.

    Trailer includes mandatory attributes for NativeMSGFileHandler.get_area_extent
    """
    trailer = {
        "15TRAILER": {
            "ImageProductionStats": {
                "ActualL15CoverageHRV": {
                    "UpperNorthLineActual": 11136,
                    "UpperWestColumnActual": 7533,
                    "UpperSouthLineActual": 8193,
                    "UpperEastColumnActual": 1966,
                    "LowerNorthLineActual": 8192,
                    "LowerWestColumnActual": 5568,
                    "LowerSouthLineActual": 1,
                    "LowerEastColumnActual": 1
                },
                "ActualScanningSummary": {
                    "ReducedScan": is_rapid_scan
                }
            }
        }
    }

    return trailer


def prepare_area_definitions(test_dict):
    """Prepare calculated and expected area definitions for equal checking."""
    earth_model = test_dict["earth_model"]
    dataset_id = test_dict["dataset_id"]
    is_full_disk = test_dict["is_full_disk"]
    is_rapid_scan = test_dict["is_rapid_scan"]
    fill_disk = test_dict["fill_disk"]
    header = create_test_header(earth_model, dataset_id, is_full_disk, is_rapid_scan)
    trailer = create_test_trailer(is_rapid_scan)
    expected_area_def = test_dict["expected_area_def"]

    with mock.patch("satpy.readers.seviri_l1b_native.fromfile") as fromfile, \
            mock.patch("satpy.readers.seviri_l1b_native.recarray2dict") as recarray2dict, \
            mock.patch("satpy.readers.seviri_l1b_native._get_array") as _get_array, \
            mock.patch(
                "satpy.readers.seviri_l1b_native.NativeMSGFileHandler._number_of_visir_channels") as _n_visir_ch, \
            mock.patch("satpy.readers.seviri_l1b_native.NativeMSGFileHandler._read_trailer"), \
            mock.patch(
                "satpy.readers.seviri_l1b_native.has_archive_header"
            ) as has_archive_header:
        has_archive_header.return_value = True
        fromfile.return_value = header
        recarray2dict.side_effect = (lambda x: x)
        _get_array.return_value = np.arange(3)
        _n_visir_ch.return_value = 11
        fh = NativeMSGFileHandler(filename=None, filename_info={}, filetype_info=None)
        fh.fill_disk = fill_disk
        fh.header = header
        fh.trailer = trailer
        fh.image_boundaries = ImageBoundaries(header, trailer, fh.mda)
        actual_area_def = fh.get_area_def(dataset_id)

    return actual_area_def, expected_area_def


@pytest.mark.parametrize(
    ("actual", "expected"),
    [
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL1_VISIR_FULLDISK)),
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL1_HRV_FULLDISK_FILL)),
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL1_VISIR_RAPIDSCAN)),
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL1_VISIR_RAPIDSCAN_FILL)),
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL1_HRV_RAPIDSCAN)),
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL1_HRV_RAPIDSCAN_FILL)),
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL1_VISIR_ROI)),
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL1_VISIR_ROI_FILL)),
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL1_HRV_ROI)),
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL1_HRV_ROI_FILL)),
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL2_VISIR_FULLDISK)),
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL2_HRV_FULLDISK_FILL)),
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL2_VISIR_RAPIDSCAN)),
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL2_VISIR_RAPIDSCAN_FILL)),
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL2_HRV_RAPIDSCAN)),
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL2_HRV_RAPIDSCAN_FILL)),
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL2_VISIR_ROI)),
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL2_VISIR_ROI_FILL)),
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL2_HRV_ROI)),
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL2_HRV_ROI_FILL)),
    ]
)
def test_area_definitions(actual, expected):
    """Test area definitions with only one area."""
    np.testing.assert_allclose(np.array(actual.area_extent),
                               np.array(expected["Area extent"]))
    assert actual.width == expected["Number of columns"]
    assert actual.height == expected["Number of rows"]
    assert actual.area_id == expected["Area ID"]


@pytest.mark.parametrize(
    ("actual", "expected"),
    [
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL1_HRV_FULLDISK)),
        (prepare_area_definitions(TEST_AREA_EXTENT_EARTHMODEL2_HRV_FULLDISK)),
    ]
)
def test_stacked_area_definitions(actual, expected):
    """Test area definitions with stacked areas."""
    np.testing.assert_allclose(np.array(actual.defs[0].area_extent),
                               np.array(expected["Area extent 0"]))
    np.testing.assert_allclose(np.array(actual.defs[1].area_extent),
                               np.array(expected["Area extent 1"]))
    assert actual.width == expected["Number of columns"]
    assert actual.height == expected["Number of rows"]
    assert actual.defs[0].area_id, expected["Area ID"]
    assert actual.defs[1].area_id, expected["Area ID"]


def prepare_is_roi(test_dict):
    """Prepare calculated and expected check for region of interest data for equal checking."""
    earth_model = 2
    dataset_id = make_dataid(name="VIS006")
    is_full_disk = test_dict["is_full_disk"]
    is_rapid_scan = test_dict["is_rapid_scan"]
    header = create_test_header(earth_model, dataset_id, is_full_disk, is_rapid_scan)
    trailer = create_test_trailer(is_rapid_scan)
    expected = test_dict["is_roi"]

    with mock.patch("satpy.readers.seviri_l1b_native.fromfile") as fromfile, \
            mock.patch("satpy.readers.seviri_l1b_native.recarray2dict") as recarray2dict, \
            mock.patch("satpy.readers.seviri_l1b_native._get_array") as _get_array, \
            mock.patch(
                "satpy.readers.seviri_l1b_native.NativeMSGFileHandler._number_of_visir_channels") as _n_visir_ch, \
            mock.patch("satpy.readers.seviri_l1b_native.NativeMSGFileHandler._read_trailer"), \
            mock.patch(
                "satpy.readers.seviri_l1b_native.has_archive_header"
            ) as has_archive_header:
        has_archive_header.return_value = True
        fromfile.return_value = header
        recarray2dict.side_effect = (lambda x: x)
        _get_array.return_value = np.arange(3)
        _n_visir_ch.return_value = 11
        fh = NativeMSGFileHandler(filename=None, filename_info={}, filetype_info=None)
        fh.header = header
        fh.trailer = trailer
        actual = fh.is_roi()

    return actual, expected


@pytest.mark.parametrize(
    ("actual", "expected"),
    [
        (prepare_is_roi(TEST_IS_ROI_FULLDISK)),
        (prepare_is_roi(TEST_IS_ROI_RAPIDSCAN)),
        (prepare_is_roi(TEST_IS_ROI_ROI)),
    ]
)
def test_is_roi(actual, expected):
    """Test if given area is of area-of-interest."""
    assert actual == expected


class TestNativeMSGFileHandler(unittest.TestCase):
    """Test the NativeMSGFileHandler."""

    def test_get_available_channels(self):
        """Test the derivation of the available channel list."""
        available_chs = get_available_channels(TEST1_HEADER_CHNLIST)
        trues = ("WV_062", "WV_073", "IR_108", "VIS006", "VIS008", "IR_120")
        for bandname in AVAILABLE_CHANNELS:
            if bandname in trues:
                assert available_chs[bandname]
            else:
                assert not available_chs[bandname]

        available_chs = get_available_channels(TEST2_HEADER_CHNLIST)
        trues = ("VIS006", "VIS008", "IR_039", "WV_062", "WV_073", "IR_087", "HRV")
        for bandname in AVAILABLE_CHANNELS:
            if bandname in trues:
                assert available_chs[bandname]
            else:
                assert not available_chs[bandname]

        available_chs = get_available_channels(TEST3_HEADER_CHNLIST)
        for bandname in AVAILABLE_CHANNELS:
            assert available_chs[bandname]


TEST_HEADER_CALIB = {
    "RadiometricProcessing": {
        "Level15ImageCalibration": {
            "CalSlope": TestFileHandlerCalibrationBase.gains_nominal,
            "CalOffset": TestFileHandlerCalibrationBase.offsets_nominal,

        },
        "MPEFCalFeedback": {
            "GSICSCalCoeff": TestFileHandlerCalibrationBase.gains_gsics,
            "GSICSOffsetCount": TestFileHandlerCalibrationBase.offsets_gsics
        }
    },
    "ImageDescription": {
        "Level15ImageProduction": {
            "PlannedChanProcessing": TestFileHandlerCalibrationBase.radiance_types
        }
    },
}


class TestNativeMSGCalibration(TestFileHandlerCalibrationBase):
    """Unit tests for calibration."""

    @pytest.fixture(name="file_handler")
    def file_handler(self):
        """Create a mocked file handler."""
        header = {
            "15_DATA_HEADER": {
                "ImageAcquisition": {
                    "PlannedAcquisitionTime": {
                        "TrueRepeatCycleStart": self.scan_time
                    }
                }
            }
        }
        trailer = {
            "15TRAILER": {
                "ImageProductionStats": {
                    "ActualScanningSummary": {
                        "ForwardScanStart": self.scan_time
                    }
                }
            }
        }
        header["15_DATA_HEADER"].update(TEST_HEADER_CALIB)
        with mock.patch("satpy.readers.seviri_l1b_native.NativeMSGFileHandler.__init__",
                        return_value=None):
            fh = NativeMSGFileHandler(filename=None, filename_info={}, filetype_info=None)
            fh.header = header
            fh.trailer = trailer
            fh.platform_id = self.platform_id
            return fh

    @pytest.mark.parametrize(
        ("channel", "calibration", "calib_mode", "use_ext_coefs"),
        [
            # VIS channel, internal coefficients
            ("VIS006", "counts", "NOMINAL", False),
            ("VIS006", "radiance", "NOMINAL", False),
            ("VIS006", "radiance", "GSICS", False),
            ("VIS006", "reflectance", "NOMINAL", False),
            # VIS channel, external coefficients (mode should have no effect)
            ("VIS006", "radiance", "GSICS", True),
            ("VIS006", "reflectance", "NOMINAL", True),
            # IR channel, internal coefficients
            ("IR_108", "counts", "NOMINAL", False),
            ("IR_108", "radiance", "NOMINAL", False),
            ("IR_108", "radiance", "GSICS", False),
            ("IR_108", "brightness_temperature", "NOMINAL", False),
            ("IR_108", "brightness_temperature", "GSICS", False),
            # IR channel, external coefficients (mode should have no effect)
            ("IR_108", "radiance", "NOMINAL", True),
            ("IR_108", "brightness_temperature", "GSICS", True),
            # HRV channel, internal coefficiens
            ("HRV", "counts", "NOMINAL", False),
            ("HRV", "radiance", "NOMINAL", False),
            ("HRV", "radiance", "GSICS", False),
            ("HRV", "reflectance", "NOMINAL", False),
            # HRV channel, external coefficients (mode should have no effect)
            ("HRV", "radiance", "GSICS", True),
            ("HRV", "reflectance", "NOMINAL", True),
        ]
    )
    def test_calibrate(
            self, file_handler, counts, channel, calibration, calib_mode,
            use_ext_coefs
    ):
        """Test the calibration."""
        external_coefs = self.external_coefs if use_ext_coefs else {}
        expected = self._get_expected(
            channel=channel,
            calibration=calibration,
            calib_mode=calib_mode,
            use_ext_coefs=use_ext_coefs
        )

        fh = file_handler
        fh.calib_mode = calib_mode
        fh.ext_calib_coefs = external_coefs

        dataset_id = make_dataid(name=channel, calibration=calibration)
        res = fh.calibrate(counts, dataset_id)
        xr.testing.assert_allclose(res, expected)


class TestNativeMSGDataset:
    """Tests for getting the dataset."""

    @pytest.fixture
    def file_handler(self):
        """Create a file handler for testing."""
        trailer = {
            "15TRAILER": {
                "ImageProductionStats": {
                    "ActualScanningSummary": {
                        "ForwardScanStart": dt.datetime(2006, 1, 1, 12, 15, 9, 304888),
                        "ForwardScanEnd": dt.datetime(2006, 1, 1, 12, 27, 9, 304888),
                        "ReducedScan": 0
                    }
                }
            }
        }
        mda = {
            "channel_list": ["VIS006", "IR_108"],
            "number_of_lines": 4,
            "number_of_columns": 4,
            "is_full_disk": True,
            "platform_name": "MSG-3",
            "offset_corrected": True,
            "projection_parameters": {
                "ssp_longitude": 0.0,
                "h": 35785831.0,
                "a": 6378169.0,
                "b": 6356583.8
            }
        }
        header = self._fake_header()
        data = self._fake_data()
        with mock.patch("satpy.readers.seviri_l1b_native.NativeMSGFileHandler.__init__",
                        return_value=None):
            fh = NativeMSGFileHandler(filename=None, filename_info={}, filetype_info=None)
            fh.header = header
            fh.trailer = trailer
            fh.mda = mda
            fh._dask_array = da.from_array(data)
            fh.platform_id = 324
            fh.fill_disk = False
            fh.calib_mode = "NOMINAL"
            fh.ext_calib_coefs = {}
            fh.include_raw_metadata = False
            fh.mda_max_array_size = 100
        return fh

    @staticmethod
    def _fake_header():
        header = {
            "15_DATA_HEADER": {
                "SatelliteStatus": {
                    "SatelliteDefinition": {
                        "NominalLongitude": 0.0
                    },
                    "Orbit": {
                        "OrbitPolynomial": ORBIT_POLYNOMIALS
                    }
                },
                "ImageAcquisition": {
                    "PlannedAcquisitionTime": {
                        "TrueRepeatCycleStart": dt.datetime(2006, 1, 1, 12, 15, 0, 0),
                        "PlannedRepeatCycleEnd": dt.datetime(2006, 1, 1, 12, 30, 0, 0),
                    }
                }
            },
        }
        header["15_DATA_HEADER"].update(TEST_HEADER_CALIB)
        return header

    @staticmethod
    def _fake_data():
        num_visir_cols = 5  # will be divided by 1.25 -> 4 columns
        visir_rec = [
            ("line_data", np.uint8, (num_visir_cols,)),
            ("acq_time", time_cds_short)
        ]
        vis006_line1 = (
            [1, 2, 3, 4, 5],  # line_data
            (1, 1000)  # acq_time (days, milliseconds)
        )
        vis006_line2 = ([6, 7, 8, 9, 10], (1, 2000))
        vis006_line3 = ([11, 12, 13, 14, 15], (1, 3000))
        vis006_line4 = ([16, 17, 18, 19, 20], (1, 4000))
        ir108_line1 = ([20, 19, 18, 17, 16], (1, 1000))
        ir108_line2 = ([15, 14, 13, 12, 11], (1, 2000))
        ir108_line3 = ([10, 9, 8, 7, 6], (1, 3000))
        ir108_line4 = ([5, 4, 3, 2, 1], (1, 4000))
        data = np.array(
            [[(vis006_line1,), (ir108_line1,)],
             [(vis006_line2,), (ir108_line2,)],
             [(vis006_line3,), (ir108_line3,)],
             [(vis006_line4,), (ir108_line4,)]],
            dtype=[("visir", visir_rec)]
        )
        return data

    def test_get_dataset(self, file_handler):
        """Test getting the dataset."""
        dataset_id = make_dataid(
            name="VIS006",
            resolution=3000,
            calibration="counts"
        )
        dataset_info = {
            "units": "1",
            "wavelength": (1, 2, 3),
            "standard_name": "counts"
        }
        xarr = file_handler.get_dataset(dataset_id, dataset_info)
        expected = self._exp_data_array()
        xr.testing.assert_equal(xarr, expected)
        assert "raw_metadata" not in xarr.attrs
        assert file_handler.start_time == dt.datetime(2006, 1, 1, 12, 15, 0)
        assert file_handler.end_time == dt.datetime(2006, 1, 1, 12, 30, 0)
        assert_attrs_equal(xarr.attrs, expected.attrs, tolerance=1e-4)

    def test_time(self, file_handler):
        """Test start/end nominal/observation time handling."""
        assert dt.datetime(2006, 1, 1, 12, 15, 9, 304888) == file_handler.observation_start_time
        assert dt.datetime(2006, 1, 1, 12, 15, ) == file_handler.start_time
        assert file_handler.start_time == file_handler.nominal_start_time

        assert dt.datetime(2006, 1, 1, 12, 27, 9, 304888) == file_handler.observation_end_time
        assert file_handler.end_time == file_handler.nominal_end_time
        assert dt.datetime(2006, 1, 1, 12, 30, ) == file_handler.end_time

    def test_repeat_cycle_duration(self, file_handler):
        """Test repeat cycle handling for FD or ReduscedScan."""
        assert 15 == file_handler._repeat_cycle_duration
        # Change the reducescan scenario to test the repeat cycle duration handling
        file_handler.trailer["15TRAILER"]["ImageProductionStats"]["ActualScanningSummary"]["ReducedScan"] = 1
        assert 5 == file_handler._repeat_cycle_duration

    @staticmethod
    def _exp_data_array():
        expected = xr.DataArray(
            np.array([[4., 32., 193., 5.],
                      [24., 112., 514., 266.],
                      [44., 192., 835., 527.],
                      [64., 273., 132., 788.]],
                     dtype=np.float32),
            dims=["y", "x"],
            attrs={
                "orbital_parameters": {
                    "satellite_actual_longitude": -3.55117540817073,
                    "satellite_actual_latitude": -0.5711243456528018,
                    "satellite_actual_altitude": 35783296.150123544,
                    "satellite_nominal_longitude": 0.0,
                    "satellite_nominal_latitude": 0.0,
                    "projection_longitude": 0.0,
                    "projection_latitude": 0.0,
                    "projection_altitude": 35785831.0
                },
                "time_parameters": {
                    "nominal_start_time": dt.datetime(2006, 1, 1, 12, 15, 0),
                    "nominal_end_time": dt.datetime(2006, 1, 1, 12, 30, 0),
                    "observation_start_time": dt.datetime(2006, 1, 1, 12, 15, 9, 304888),
                    "observation_end_time": dt.datetime(2006, 1, 1, 12, 27, 9, 304888),
                },
                "georef_offset_corrected": True,
                "platform_name": "MSG-3",
                "sensor": "seviri",
                "units": "1",
                "wavelength": (1, 2, 3),
                "standard_name": "counts",
            }
        )
        expected["acq_time"] = ("y", [np.datetime64("1958-01-02 00:00:01").astype("datetime64[ns]"),
                                      np.datetime64("1958-01-02 00:00:02").astype("datetime64[ns]"),
                                      np.datetime64("1958-01-02 00:00:03").astype("datetime64[ns]"),
                                      np.datetime64("1958-01-02 00:00:04").astype("datetime64[ns]")])
        return expected

    def test_get_dataset_with_raw_metadata(self, file_handler):
        """Test provision of raw metadata."""
        file_handler.include_raw_metadata = True
        dataset_id = make_dataid(
            name="VIS006",
            resolution=3000,
            calibration="counts"
        )
        dataset_info = {
            "units": "1",
            "wavelength": (1, 2, 3),
            "standard_name": "counts"
        }
        xarr = file_handler.get_dataset(dataset_id, dataset_info)
        assert "raw_metadata" in xarr.attrs

    def test_satpos_no_valid_orbit_polynomial(self, file_handler):
        """Test satellite position if there is no valid orbit polynomial."""
        file_handler.header["15_DATA_HEADER"]["SatelliteStatus"][
            "Orbit"]["OrbitPolynomial"] = ORBIT_POLYNOMIALS_INVALID
        dataset_id = make_dataid(
            name="VIS006",
            resolution=3000,
            calibration="counts"
        )
        dataset_info = {
            "units": "1",
            "wavelength": (1, 2, 3),
            "standard_name": "counts"
        }
        with pytest.warns(UserWarning, match="No orbit polynomial"):
            xarr = file_handler.get_dataset(dataset_id, dataset_info)
        assert "satellite_actual_longitude" not in xarr.attrs[
            "orbital_parameters"]


class TestNativeMSGPadder(unittest.TestCase):
    """Test Padder of the native l1b seviri reader."""

    @staticmethod
    def prepare_padder(test_dict):
        """Initialize Padder and pad test data."""
        dataset_id = test_dict["dataset_id"]
        img_bounds = test_dict["img_bounds"]
        is_full_disk = test_dict["is_full_disk"]
        dataset = test_dict["dataset"]
        final_shape = test_dict["final_shape"]
        expected_padded_data = test_dict["expected_padded_data"]

        padder = Padder(dataset_id, img_bounds, is_full_disk)
        padder._final_shape = final_shape
        calc_padded_data = padder.pad_data(dataset)

        return calc_padded_data, expected_padded_data

    def test_padder_rss_roi(self):
        """Test padder for RSS and ROI data (applies to both VISIR and HRV)."""
        calculated, expected = self.prepare_padder(TEST_PADDER_RSS_ROI)
        np.testing.assert_array_equal(calculated, expected)

    def test_padder_fes_hrv(self):
        """Test padder for FES HRV data."""
        calculated, expected = self.prepare_padder(TEST_PADDER_FES_HRV)
        np.testing.assert_array_equal(calculated, expected)


class TestNativeMSGFilenames:
    """Test identification of Native format filenames."""

    @pytest.fixture
    def reader(self):
        """Return reader for SEVIRI Native format."""
        from satpy._config import config_search_paths
        from satpy.readers.core.loading import load_reader

        reader_configs = config_search_paths(
            os.path.join("readers", "seviri_l1b_native.yaml"))
        reader = load_reader(reader_configs)
        return reader

    def test_file_pattern(self, reader):
        """Test file pattern matching."""
        filenames = [
            # Valid
            "MSG2-SEVI-MSG15-0100-NA-20080219094242.289000000Z",
            "MSG2-SEVI-MSG15-0201-NA-20080219094242.289000000Z",
            "MSG2-SEVI-MSG15-0301-NA-20080219094242.289000000Z-123456.nat",
            "MSG2-SEVI-MSG15-0401-NA-20080219094242.289000000Z-20201231181545-123456.nat",
            # Invalid
            "MSG2-SEVI-MSG15-010-NA-20080219094242.289000000Z",
        ]
        files = reader.select_files_from_pathnames(filenames)
        assert len(files) == 4


@pytest.mark.parametrize(
    ("file_content", "exp_header_size"),
    [
        (ASCII_STARTSWITH, 450400),  # with ascii header
        (b"foobar", 445286),  # without ascii header
    ]
)
def test_header_type(file_content, exp_header_size):
    """Test identification of the file header type."""
    header = create_test_header(
        dataset_id=make_dataid(name="VIS006", resolution=3000),
        earth_model=1,
        is_full_disk=True,
        is_rapid_scan=0
    )
    if file_content == b"foobar":
        header.pop("15_SECONDARY_PRODUCT_HEADER")
    with mock.patch("satpy.readers.seviri_l1b_native.fromfile") as fromfile, \
            mock.patch("satpy.readers.seviri_l1b_native.recarray2dict") as recarray2dict, \
            mock.patch("satpy.readers.seviri_l1b_native._get_array") as _get_array, \
            mock.patch(
                "satpy.readers.seviri_l1b_native.NativeMSGFileHandler._number_of_visir_channels") as _n_visir_ch, \
            mock.patch("satpy.readers.seviri_l1b_native.NativeMSGFileHandler._read_trailer"), \
            mock.patch("satpy.readers.seviri_l1b_native.generic_open", mock.mock_open(read_data=file_content)):
        fromfile.return_value = header
        recarray2dict.side_effect = (lambda x: x)
        _get_array.return_value = np.arange(3)
        _n_visir_ch.return_value = 11
        fh = NativeMSGFileHandler(filename=None, filename_info={}, filetype_info=None)
        assert fh.header_type.itemsize == exp_header_size
        assert "15_SECONDARY_PRODUCT_HEADER" in fh.header


def test_header_warning():
    """Test warning is raised for NOK quality flag."""
    header_good = create_test_header(
        dataset_id=make_dataid(name="VIS006", resolution=3000),
        earth_model=1,
        is_full_disk=True,
        is_rapid_scan=0,
        good_qual="OK"
    )
    header_bad = create_test_header(
        dataset_id=make_dataid(name="VIS006", resolution=3000),
        earth_model=1,
        is_full_disk=True,
        is_rapid_scan=0,
        good_qual="NOK"
    )

    with mock.patch("satpy.readers.seviri_l1b_native.fromfile") as fromfile, \
            mock.patch("satpy.readers.seviri_l1b_native.recarray2dict") as recarray2dict, \
            mock.patch("satpy.readers.seviri_l1b_native._get_array") as _get_array, \
            mock.patch(
                "satpy.readers.seviri_l1b_native.NativeMSGFileHandler._number_of_visir_channels") as _n_visir_ch, \
            mock.patch("satpy.readers.seviri_l1b_native.NativeMSGFileHandler._read_trailer"), \
            mock.patch("satpy.readers.seviri_l1b_native.generic_open", mock.mock_open(read_data=ASCII_STARTSWITH)):
        recarray2dict.side_effect = (lambda x: x)
        _get_array.return_value = np.arange(3)

        exp_warning = "The quality flag for this file indicates not OK. Use this data with caution!"

        fromfile.return_value = header_good
        _n_visir_ch.return_value = 11

        with warnings.catch_warnings():
            warnings.simplefilter("error")
            NativeMSGFileHandler(filename=None, filename_info={}, filetype_info=None)

        fromfile.return_value = header_bad
        with pytest.warns(UserWarning, match=exp_warning):
            NativeMSGFileHandler(filename=None, filename_info={}, filetype_info=None)

        # check that without Main Header the code doesn't crash
        header_missing = header_good.copy()
        header_missing.pop("15_MAIN_PRODUCT_HEADER")
        fromfile.return_value = header_missing
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            NativeMSGFileHandler("myfile", {}, None)


@pytest.mark.parametrize(
    ("starts_with", "expected"),
    [
        (ASCII_STARTSWITH, True),
        (b"this_shall_fail", False)
    ]
)
def test_has_archive_header(starts_with, expected):
    """Test if the file includes an ASCII archive header."""
    with mock.patch("satpy.readers.seviri_l1b_native.generic_open", mock.mock_open(read_data=starts_with)):
        actual = has_archive_header("filename")
    assert actual == expected


def test_read_header():
    """Test that reading header returns the header correctly converted to a dictionary."""
    keys = ("SatelliteId", "NominalLongitude", "SatelliteStatus")
    values = (324, 0.0, 1)
    expected = dict(zip(keys, values))

    types = (np.uint16, np.float32, np.uint8)
    dtypes = np.dtype([(k, t) for k, t in zip(keys, types)])
    hdr_data = np.array([values], dtype=dtypes)

    with mock.patch("satpy.readers.seviri_l1b_native.fromfile") as fromfile:
        fromfile.return_value = hdr_data
        actual = recarray2dict(hdr_data)
    assert actual == expected


@pytest.fixture(scope="session")
def tmp_seviri_nat_filename(session_tmp_path):
    """Create a fully-qualified filename for a seviri native format file."""
    full_file_path = session_tmp_path / "MSG4-SEVI-MSG15-0100-NA-20210528075743.722000000Z-N.nat"
    create_physical_seviri_native_file(full_file_path)
    return full_file_path


@pytest.fixture(scope="session")
def compressed_seviri_native_file(tmp_seviri_nat_filename, session_tmp_path):
    """Return the fsspec path to the given seviri native file inside a zip file."""
    zip_full_path = session_tmp_path / "test_seviri_native.zip"
    with zipfile.ZipFile(zip_full_path, mode="w") as archive:
        archive.write(tmp_seviri_nat_filename, os.path.basename(tmp_seviri_nat_filename))
    return f"zip://*.nat::file://{zip_full_path.as_posix()}"


@pytest.mark.parametrize(("full_path"), [
    lf("tmp_seviri_nat_filename"),
    lf("compressed_seviri_native_file")
])
def test_read_physical_seviri_nat_file(full_path):
    """Test that the physical seviri native file can be read successfully, in case of both a plain and a zip file.

    Note:
        The purpose of this function is not to fully test the properties of the scene. It only provides a test for
        reading a physical file from disk.
    """
    scene = scene_from_physical_seviri_nat_file(full_path)

    assert scene.sensor_names == {"seviri"}
    assert len(scene.available_dataset_ids()) == 36
    assert set(scene.available_dataset_names()) == set(CHANNEL_INDEX_LIST)

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning, message="No orbit polynomial valid")
        scene.load(["VIS006"])
        assert scene["VIS006"].dtype == np.float32
        assert scene["VIS006"].values.dtype == np.float32
        assert scene["VIS006"].shape == (3712, 3712)
        assert isinstance(scene["VIS006"], xr.core.dataarray.DataArray)


def scene_from_physical_seviri_nat_file(filename):
    """Generate a Scene object from the given seviri native file."""
    return Scene([filename], reader="seviri_l1b_native", reader_kwargs={"fill_disk": True})


def create_physical_seviri_native_file(seviri_nat_full_file_path):
    """Create a physical seviri native file on disk."""
    header_type, header_null = generate_seviri_native_null_header()
    amend_seviri_native_null_header(header_null)
    append_data_and_trailer_content_to_seviri_native_header(seviri_nat_full_file_path, header_null)


def generate_seviri_native_null_header():
    """Generate the header of the seviri native format which is filled with zeros, hence the term null!"""
    header_type = Msg15NativeHeaderRecord().get(True)
    null_header = np.zeros(header_type.shape, dtype=header_type).reshape(1, )
    return header_type, null_header


def amend_seviri_native_null_header(hdr_null_numpy):
    """Amend the given null header so that the ``seviri_l1b_native`` reader can properly parse it.

    This is achieved by setting values for the bare minimum number of header fields so that the reader can make sense of
    the given header. This function relies on a number of auxiliary functions all of which are enclosed in the body of
    the present function.

    Note:
        The naming scheme of the auxiliary functions is as follows: ``_amend_<KEY1>__<KEY2>__...``, where <KEY{n}>
        corresponds to keys in the header when it is represented as a dictionary, i.e. when calling ``recarray2dict()``
        on the given header array.

        For example, ``_amend_15_DATA_HEADER__SatelliteStatus__SatelliteDefinition__SatelliteId()`` corresponds to an
        auxiliary function which manipulates the following entry:
        ``hdr_null_numpy_as_dict["15_DATA_HEADER"]["SatelliteStatus"]["SatelliteDefinition"]["SatelliteId"]``
    """

    def _amend_15_MAIN_PRODUCT_HEADER():
        hdr_null_numpy[0][0][0] = (b"FormatName                  : ", b"NATIVE\n")

    def _amend_15_SECONDARY_PRODUCT_HEADER():
        hdr_null_numpy[0][1][9] = (b"SelectedBandIDs", b"XXXXXXXXXXX-")
        hdr_null_numpy[0][1][10] = (b"SouthLineSelectedRectangle", b"3360")
        hdr_null_numpy[0][1][11] = (b"NorthLineSelectedRectangle", b"3373")
        hdr_null_numpy[0][1][12] = (b"EastColumnSelectedRectangle", b"1714")
        hdr_null_numpy[0][1][13] = (b"WestColumnSelectedRectangle", b"1729")
        hdr_null_numpy[0][1][14] = (b"NumberLinesVISIR", b"14")
        hdr_null_numpy[0][1][15] = (b"NumberColumnsVISIR", b"16")
        hdr_null_numpy[0][1][16] = (b"NumberLinesHRV", b"42")
        hdr_null_numpy[0][1][17] = (b"NumberColumnsHRV", b"48")

    def _amend_GP_PK_SH1__PacketTime():
        hdr_null_numpy[0][3][5] = (23158, 27921912)

    def _amend_15_DATA_HEADER__SatelliteStatus__SatelliteDefinition__SatelliteId():
        hdr_null_numpy[0][4][1][0][0] = 324

    def _amend_15_DATA_HEADER__GeometricProcessing__EarthModel():
        hdr_null_numpy[0][4][6][1] = (2, 6378.169, 6356.5838, 6356.5838)

    def _amend_15_DATA_HEADER__ImageAcquisition__PlannedAcquisitionTime():
        hdr_null_numpy[0][4][2][0] = (
            (23158, 27911177, 286, 223),
            (23158, 28663675, 401, 687),
            (23158, 28810078, 157, 663)
        )

    # Apply all the header amendments
    _amend_15_MAIN_PRODUCT_HEADER()
    _amend_15_SECONDARY_PRODUCT_HEADER()
    _amend_GP_PK_SH1__PacketTime()
    _amend_15_DATA_HEADER__SatelliteStatus__SatelliteDefinition__SatelliteId()
    _amend_15_DATA_HEADER__GeometricProcessing__EarthModel()
    _amend_15_DATA_HEADER__ImageAcquisition__PlannedAcquisitionTime()


def append_data_and_trailer_content_to_seviri_native_header(filename, hdr_null_numpy):
    """Generate the data and trailer part (null content) of the file and appends them to the null header.

    The data and trailer are also null and appending them to the header results in a complete seviri nat file.
    """
    # size of different parts of the seviri native file in bytes
    size = {"header_with_archive": 450400, "data": 13090, "trailer": 380363}
    zero_bytes = bytearray(size["data"] + size["trailer"])
    bytes_data = bytes(zero_bytes)

    hdr_null_numpy.tofile(filename)
    with open(filename, "ab") as f:
        f.write(bytes_data)



# Copyright (c) 2022 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.
r"""Base class used for the MTG Lighting Imager netCDF4 readers.

The base LI reader class supports generating the available datasets
programmatically: to achieve this, each LI product type should provide a
\"file description\" which is itself retrieved directly from the YAML
configuration file for the reader of interest, as a custom
``file_desc`` entry inside the \'file_type\' section
corresponding to that product type.

Each of the ``file_desc`` entry describes what are the
variables that are available into that product that should be used to
register the available satpy datasets.

Each of those description entries may contain the following elements:

- **product_type** \[required\]:

  Indicate the processing_level / product_type name to use internally for
  that type of product file. This should correspond to the
  ``{processing_level}-{product_type}`` part of the full
  file_pattern.

- **search_paths** \[optional\]:

  A list of the possible paths that should be prefixed to a given variable
  name when searching for that variable in the NetCDF file to register a
  dataset on it. The list is given in priority order. If no search path is
  provided (or an empty array is provided) then the variables will only be
  searched directly in the root group of the NetCDF structure.

- **swath_coordinates** \[required\]:

  The LI reader will use a ``SwathDefinition`` object to define
  the area/coordinates of each of the provided datasets depending on the
  content of this entry. The user can either:

    - Specify a
      ``swath_coordinates`` entry directly with
      ``latitude`` and ``longitude`` entries, in which
      case, the datasets that will match one of the
      ``'variable_patterns'`` provided will use those lat/lon
      variables as coordinate providers.
    - Specify a
      ``swath_coordinates`` entry directly with
      ``projection``, ``azimuth`` and
      ``elevation`` entries instead, in which case, the reader will
      first use the variables pointed by those 3 entries compute the
      corresponding latitude/longitude data from the scan angles contained in
      the product file. And then, continue with assigned those lat/lon
      datasets as coordinates for datasets that will match one of the
      ``variable_patterns`` provided.

  **Note:** It is acceptable to specify an empty array for the list of
  ``variable_patterns``, in this case, the swath coordinates will not be
  assigned to any dataset.

- **sectors** \[optional\]:

  The custom dataset description mechanism makes a distinction between
  \"ordinary\" variables which should be used to create a \"single
  dataset\" and \"sectored variables\" which will be found per sector and
  will thus be used to generate as many datasets as there are sectors (see below). So
  this entry is used to specify the list of sector names there should be
  available in the NetCDF structure.

- **sector_variables** \[optional\]:

  This entry is used to provide a list of the variables that are available
  **per sector** in the NetCDF file. Thus, assuming the
  ``sectors`` entry is set to the standard list ``['north',
  'east', 'south', 'west']``, 4 separated datasets will
  be registered for each variable listed here (using the conventional
  suffix ``"{sector_name}_sector"``)

- **variables** \[optional\]:

  This entry is used to provide a list of \"ordinary variables\" (ie.
  variables that are not available **per sector**). Each of those
  variables will be used to register one dataset.

  **Note:** A single product may provide both the \"variables\" and the
  \"sector_variables\" at the same time (as this is the case for LI LEF
  for instance)

- **variable_transforms** \[optional\]:

  This entry is may be used to provide specific additional entries **per
  variable name** (ie. will apply to both in sector or out of sector
  variables) that should be added to the dataset infos when registering a
  dataset with that variable. While any kind of info could be added this
  way to the final dataset infos, we are currently using the entry mainly
  to provide our LI reader with the following traits which
  will then be used to \"transform\" the data of the dataset as requested
  on loading:

    - ``broadcast_to``: if this extra info is found in a
      dataset_info on dataset loading, then the initial data array will be
      broadcast to the shape of the variable found under the variable path
      specified as value for that entry. Note that, if the pattern
      ``{sector_name}`` if found in this entry value, then the
      reader will assume that we are writing a dataset from an in sector
      variable, and use the current sector name to find the appropriate
      alternate variable that will be used as reference to broadcast the
      current variable data.

    - ``seconds_to_datetime``: This transformation is used to
      internally convert variables provided as float values to the
      ``np.datetime64`` data type. The value specified for this
      entry should be the reference epoch time used as offsets for the
      elapsed seconds when converting the data.

    - ``seconds_to_timedelta``: This transformation is used to
      internally convert variables (assumed to use a \"second\" unit)
      provided as float values to the ``np.timedelta64`` data
      type. This entry should be set to ``true`` to activate this
      transform. During the conversion, we internally use a
      nanosecond resolution on the input floating point second values.

    - ``milliseconds_to_timedelta``: Same kind of transformation
      as ``seconds_to_timedelta`` except that the source data is
      assumed to contain millisecond float values.

    - ``accumulate_index_offset``: if this extra info is found in
      a ``dataset_info`` on dataset loading, then we will consider that the
      dataset currently being generated is an array of indices inside the
      variable pointed by the path provided as value for that entry. Note
      that the same usage of the pattern ``{sector_name}``
      mentioned for the entry \"broadcast_to\" will also apply here. This
      behavior is useful when multiple input files are loaded together in a
      single satpy scene, in which case, the variables from each files will
      be concatenated to produce a single dataset for each variable, and
      thus the need to correct the reported indices accordingly.

      An example of usage of this entry is as follows:

      ..  code-block:: yaml

          variable_transforms:
            integration_frame_index:
              accumulate_index_offset: "{sector_name}/exposure_time"



      In the example above the integration_frame_index from each sector (i.e.
      optical channel) provides a list of indices in the corresponding
      exposure_time array from that same sector. The final indices will thus
      correctly take into account that the final exposure_time array contains
      all the values concatenated from all the input files in the scene.

    - ``use_rescaling``: By default, we currently
      apply variable rescaling as soon as we find one (or more) of the
      attributes named ``'scale_factor'``,
      ``'scaling_factor'`` or ``'add_offset'`` in
      the source netcdf variable. This automatic transformation can be
      disabled for a given variable specifying a value of false for
      this extra info element, for instance:

      ..  code-block:: yaml

          variable_transforms:
            latitude:
              use_rescaling: false


      **Note:** We are currently not disabling rescaling for any dataset, so
      that entry is not used in the current version of the YAML config files
      for the LI readers.

"""

import logging
import re

import dask.array as da
import netCDF4
import numpy as np
import xarray as xr
from pyproj import Proj

from satpy.readers.core.fci import platform_name_translate
from satpy.readers.core.netcdf import NetCDF4FsspecFileHandler

logger = logging.getLogger(__name__)


class LINCFileHandler(NetCDF4FsspecFileHandler):
    """Base class used as parent for the concrete LI reader classes."""

    def __init__(self, filename, filename_info, filetype_info, cache_handle=True):
        """Initialize LINCFileHandler."""
        super().__init__(filename, filename_info, filetype_info,
                         cache_var_size=10000,
                         cache_handle=cache_handle
                         )

        # decode_times should be disabled for xr.open_dataset access (cache_handle=False):
        # Note: the default dict assignment is need to avoid error when using the fake
        # netcdf4 file handler in mock unit tests:
        self._xarray_kwargs = getattr(self, "_xarray_kwargs", {})
        self._xarray_kwargs["decode_times"] = False
        self._xarray_kwargs["mask_and_scale"] = False

        # Processing level that should be set by derived classes.
        self.processing_level = filetype_info.get("processing_level", "L0")

        # This class will only provide support for the LI sensor:
        self.sensors = {"li"}

        # Set of dataset names explicitly provided by this file handler:
        # This set is required to filter the retrieval of datasets later in the
        # get_dataset() method, for instance when building a Scene from multiple
        # different product files (example: 1 L1B BCK file + 1 L1B LE file):
        # the dataset loading mechanism will still request from the LE specific
        # file handler if it can load a 'timestamp_vis_08_north_sector' dataset
        # for instance.
        # And when concatenating multiple BCK files into a single scene, usually
        # only one of the file handler will be able to load a specific timestamp.
        # => We could recompute the availability of a dataset from the provided
        # ds_info in get_dataset(), but it seems a better/easier solution to just
        # cache the set of available dataset names as generated in 'available_datasets()'
        # directly here:
        self.provided_datasets = set()

        self.ds_desc = filetype_info["file_desc"]
        # Store the extra infos available on specific variables:
        # Write the correct product type here:
        self.product_type = self.ds_desc["product_type"]
        logger.debug("Product type is: %s", self.product_type)

        self.variable_transforms = self.ds_desc.get("variable_transforms", {})

        # Store the pattern for the default swath coordinates:
        # Note that we should always have this swath coordinates entry now:
        self.swath_coordinates = self.ds_desc.get("swath_coordinates", {})
        patterns = self.swath_coordinates.get("variable_patterns", [])
        self.swath_coordinates["patterns"] = [re.compile(pstr) for pstr in patterns]

        # check if the current product is in an accumulation grid
        self.prod_in_accumulation_grid = self.is_prod_in_accumulation_grid()

        # list of paths where we should be looking for data when trying to retrieve
        # "measured variables" from the netcdf file attached to this file handler.
        self.search_paths = None

        # Storage for the registered datasets provided in this file handler: will only be
        # initialized once in `register_available_datasets()`
        self.dataset_infos = None

        # Storage for the current ds_infos in use in a call to get_dataset()
        self.current_ds_info = None

        # Ordered list of transform operations supported in this file handler:
        # those transforms are applied if requested in the 'apply_transforms' method below
        self.transform_names = ["use_rescaling", "seconds_to_timedelta", "milliseconds_to_timedelta",
                                "seconds_to_datetime", "broadcast_to", "accumulate_index_offset"]

        # store internal variables
        self.internal_variables = {}

        # We register all the available datasets on creation:
        self.register_available_datasets()

    @property
    def start_time(self):
        """Get the start time."""
        return self.filename_info["start_time"]

    @property
    def end_time(self):
        """Get the end time."""
        return self.filename_info["end_time"]

    @property
    def sensor_names(self):
        """List of sensors represented in this file."""
        return self.sensors

    def is_prod_in_accumulation_grid(self):
        """Check if the current product is an accumulated product in geos grid."""
        in_grid = self.swath_coordinates.get("projection", None) == "mtg_geos_projection"
        return in_grid

    def get_latlon_names(self):
        """Retrieve the user specified names for latitude/longitude coordinates.

        Use default 'latitude' / 'longitude' if not specified.
        """
        lon_name = self.swath_coordinates.setdefault("longitude", "longitude")
        lat_name = self.swath_coordinates.setdefault("latitude", "latitude")
        return lat_name, lon_name

    def get_projection_config(self):
        """Retrieve the projection configuration details."""
        # We retrieve the projection variable name directly from our swath settings:
        proj_var = self.swath_coordinates["projection"]
        geos_proj = self.get_measured_variable(proj_var, fill_value=None)
        # cast projection attributes to float/str:
        major_axis = float(geos_proj.attrs["semi_major_axis"])
        # TODO reinstate reading from file when test data issue is fixed
        point_height = 35786400.0  # float(geos_proj.attrs["perspective_point_height"])
        inv_flattening = float(geos_proj.attrs["inverse_flattening"])
        lon_0 = float(geos_proj.attrs["longitude_of_projection_origin"])
        sweep = str(geos_proj.attrs["sweep_angle_axis"])

        # use a (semi-major axis) and rf (reverse flattening) to define ellipsoid as recommended by EUM
        proj_dict = {"a": major_axis,
                     "lon_0": lon_0,
                     "h": point_height,
                     "rf": inv_flattening,
                     "proj": "geos",
                     "units": "m",
                     "sweep": sweep}

        return proj_dict

    def get_daskified_lon_lat(self, proj_dict):
        """Get daskified lon and lat array using map_blocks."""
        # Get our azimuth/elevation arrays,
        azimuth = self.get_measured_variable(self.swath_coordinates["azimuth"])
        azimuth = self.apply_use_rescaling(azimuth)

        elevation = self.get_measured_variable(self.swath_coordinates["elevation"])
        elevation = self.apply_use_rescaling(elevation)

        # Daskify inverse projection computation:
        lon, lat = da.map_blocks(self.inverse_projection, azimuth, elevation, proj_dict,
                                 chunks=(2, azimuth.shape[0]),
                                 meta=np.array((), dtype=azimuth.dtype),
                                 dtype=azimuth.dtype,
                                 )
        return lon, lat

    def generate_coords_from_scan_angles(self):
        """Generate the latitude/longitude coordinates from the scan azimuth and elevation angles."""
        proj_cfg = self.get_projection_config()
        lon, lat = self.get_daskified_lon_lat(proj_cfg)

        # Retrieve the names we should use for the generated lat/lon datasets:
        lat_name, lon_name = self.get_latlon_names()

        # Finally, we should store those arrays as internal variables for later retrieval as
        # standard datasets:
        self.internal_variables[lon_name] = xr.DataArray(
            da.asarray(lon), dims=["y"], attrs={"standard_name": "longitude"}).astype(np.float32)
        self.internal_variables[lat_name] = xr.DataArray(
            da.asarray(lat), dims=["y"], attrs={"standard_name": "latitude"}).astype(np.float32)

    def inverse_projection(self, azimuth, elevation, proj_dict):
        """Compute inverse projection."""
        # Initialise Proj object:
        projection = Proj(proj_dict)

        # Retrieve the point height from the projection config:
        point_height = proj_dict["h"]

        # Convert scan angles to projection coordinates by multiplying with perspective point height
        azimuth = azimuth.values * point_height
        elevation = elevation.values * point_height

        # In the MTG world, azimuth is defined as positive towards west, while proj expects it positive towards east
        azimuth *= -1

        lon, lat = projection(azimuth, elevation, inverse=True)

        return np.stack([lon.astype(azimuth.dtype), lat.astype(elevation.dtype)])

    def register_coords_from_scan_angles(self):
        """Register lat lon datasets in this reader."""
        lat_name, lon_name = self.get_latlon_names()
        self.register_dataset(lon_name)
        self.register_dataset(lat_name)

    def variable_path_exists(self, var_path):
        """Check if a given variable path is available in the underlying netCDF file.

        All we really need to do here is to access the file_content dictionary and
        check if we have a variable under that var_path key.
        """
        # but we ignore attributes: or sub properties:
        if var_path.startswith("/attr") or var_path.endswith(("/dtype", "/shape", "/dimensions")):
            return False

        # Check if the path is found:
        if var_path in self.file_content:
            # This is only a valid variable if it is not a netcdf group:
            return not isinstance(self.file_content[var_path], netCDF4.Group)

        # Var path not in file_content:
        return False

    def get_first_valid_variable(self, var_paths):
        """Select the first valid path for a variable from the given input list and returns the data."""
        for vpath in var_paths:
            if self.variable_path_exists(vpath):
                return self[vpath]

        # We could not find a variable with that path, this might be an error:
        raise KeyError(f"Could not find variable with paths: {var_paths}")

    def get_measured_variable(self, var_paths, fill_value=np.nan):
        """Retrieve a measured variable path taking into account the potential old data formatting schema.

        And also replace the missing values with the provided fill_value (except if this is explicitly
        set to None).
        Also, if a slice index is provided, only that slice of the array (on the axis=0) is retrieved
        (before filling the missing values).
        """
        # convert the var_paths to a list in case it is a single string:
        if isinstance(var_paths, str):
            var_paths = [var_paths]

        # then we may return one of the internal variables:
        # We only really need to check the first variable name in the list below:
        # it doesn't really make sense to mix internal variables and multi var
        # names anyway
        for vname, arr in self.internal_variables.items():
            if var_paths[0].endswith(vname):
                return arr

        # Get the search paths from our dataset descriptions:
        all_var_paths = self.get_variable_search_paths(var_paths)

        arr = self.get_first_valid_variable(all_var_paths)

        # Also handle fill value here (but only if it is not None, so that we can still bypass this
        # step if needed)
        arr = self.apply_fill_value(arr, fill_value)
        return arr

    def apply_fill_value(self, arr, fill_value):
        """Apply fill values, unless it is None and when _FillValue is provided in the array attributes."""
        if fill_value is not None and arr.attrs.get("_FillValue") is not None:
            if np.isnan(fill_value):
                fill_value = np.float32(np.nan)
            arr = arr.where(arr != arr.attrs.get("_FillValue"), fill_value)
        return arr

    def get_variable_search_paths(self, var_paths):
        """Get the search paths from the dataset descriptions."""
        if len(self.search_paths) == 0:
            all_var_paths = var_paths
        else:
            all_var_paths = [f"{folder}/{var_path}"
                             for folder in self.search_paths
                             for var_path in var_paths]
        return all_var_paths

    def add_provided_dataset(self, ds_infos):
        """Add a provided dataset to our internal list."""
        # Check if we have extra infos for that variable:
        # Note that if available we should use the alias name instead here:
        vname = ds_infos["alias_name"] if "alias_name" in ds_infos else ds_infos["variable_name"]
        self.check_variable_extra_info(ds_infos, vname)

        # We check here if we should include the default coordinates on that dataset:
        if self.swath_coordinates is not None and "coordinates" not in ds_infos:

            # Check if the variable corresponding to this dataset will match one of the valid patterns
            # for the swath usage:
            if any([p.search(vname) is not None for p in self.swath_coordinates["patterns"]]):

                # Get the target coordinate names, applying the sector name as needed:
                lat_coord_name, lon_coord_name = self.get_coordinate_names(ds_infos)

                # Ensure we do not try to add the coordinates on the coordinates themself:
                dname = ds_infos["name"]
                if dname != lat_coord_name and dname != lon_coord_name:
                    ds_infos["coordinates"] = [lon_coord_name, lat_coord_name]
        self.dataset_infos.append(ds_infos)
        self.provided_datasets.add(ds_infos["name"])

    def check_variable_extra_info(self, ds_infos, vname):
        """Check if we have extra infos for that variable."""
        if vname in self.variable_transforms:
            extras = self.variable_transforms[vname]

            # extend the ds_infos:
            ds_infos.update(extras)

    def get_coordinate_names(self, ds_infos):
        """Get the target coordinate names, applying the sector name as needed."""
        lat_coord_name, lon_coord_name = self.get_latlon_names()
        if "sector_name" in ds_infos:
            sname = ds_infos["sector_name"]
            lat_coord_name = lat_coord_name.replace("{sector_name}", sname)
            lon_coord_name = lon_coord_name.replace("{sector_name}", sname)
        return lat_coord_name, lon_coord_name

    def get_dataset_infos(self, dname):
        """Retrieve the dataset infos corresponding to one of the registered datasets."""
        for dsinfos in self.dataset_infos:
            if dsinfos["name"] == dname:
                return dsinfos

        # nothing found.
        return None

    def register_dataset(self, var_name, oc_name=None):
        """Register a simple dataset given name elements."""
        # generate our default dataset name:

        ds_name = var_name if oc_name is None else f"{var_name}_{oc_name}_sector"

        platform = self.filename_info["mission_prefix"] + "I" + self.filename_info["spacecraft_id"]

        ds_info = {
            "name": ds_name,
            "variable_name": var_name,
            "sensor": "li",
            "platform_name": platform_name_translate[platform],
            "file_type": self.filetype_info["file_type"]
        }

        # add the sector name:
        if oc_name is not None:
            ds_info["sector_name"] = oc_name

        self.add_provided_dataset(ds_info)

    def register_available_datasets(self):
        """Register all the available dataset that should be made available from this file handler."""
        if self.dataset_infos is not None:
            return

        # Otherwise, we need to perform the registration:
        self.dataset_infos = []

        # Assign the search paths for this product type:
        self.search_paths = self.ds_desc.get("search_paths", [])

        # Register our coordinates from azimuth/elevation data
        # if the product is accumulated
        if self.prod_in_accumulation_grid:
            self.register_coords_from_scan_angles()

        # First we check if we have support for sectors for this product:
        self.register_sector_datasets()

        # Retrieve the list of "raw" (ie not in sectors) variables provided in this description:
        self.register_variable_datasets()

        logger.debug("Adding %d datasets for %s input product.",
                     len(self.dataset_infos), self.product_type)

    def register_variable_datasets(self):
        """Register all the available raw (i.e. not in sectors)."""
        if "variables" in self.ds_desc:
            all_vars = self.ds_desc["variables"]
            # No sector to handle so we write simple datasets from the variables:
            for var_name in all_vars:
                self.register_dataset(var_name)

    def register_sector_datasets(self):
        """Register all the available sector datasets."""
        if "sectors" in self.ds_desc:
            sectors = self.ds_desc["sectors"]
            sector_vars = self.ds_desc["sector_variables"]
            # We should generate the datasets per sector:
            for oc_name in sectors:
                for var_name in sector_vars:
                    self.register_dataset(var_name, oc_name)

    def available_datasets(self, configured_datasets=None):
        """Determine automatically the datasets provided by this file.

        Uses a per product type dataset registration mechanism using the dataset descriptions declared in the reader
        construction above.
        """
        # pass along existing datasets
        for is_avail, ds_info in (configured_datasets or []):
            yield is_avail, ds_info

        for ds_info in self.dataset_infos:
            yield True, ds_info

    def apply_use_rescaling(self, data_array, ds_info=None):
        """Apply the use_rescaling transform on a given array."""
        # Here we should apply the rescaling except if it is explicitly requested not to rescale
        if ds_info is not None and ds_info.get("use_rescaling", True) is not True:
            return data_array

        # Check if we have the scaling elements:
        attribs = data_array.attrs
        if "scale_factor" in attribs or "scaling_factor" in attribs or "add_offset" in attribs:
            # TODO remove scaling_factor fallback after issue in NetCDF is fixed
            scale_factor = attribs.setdefault("scale_factor", attribs.get("scaling_factor", 1))
            add_offset = attribs.setdefault("add_offset", 0)
            data_array = (data_array * scale_factor) + add_offset
            # rescale the valid range accordingly
            if "valid_range" in attribs.keys():
                attribs["valid_range"] = attribs["valid_range"] * scale_factor + add_offset

        data_array.attrs.update(attribs)

        return data_array

    def apply_broadcast_to(self, data_array, ds_info):
        """Apply the broadcast_to transform on a given array."""
        ref_var = self.get_transform_reference("broadcast_to", ds_info)

        logger.debug("Broascasting %s to shape %s", ds_info["name"], ref_var.shape)
        new_array = da.broadcast_to(data_array, ref_var.shape)
        dims = data_array.dims if data_array.ndim > 0 else ("y",)
        data_array = xr.DataArray(new_array, coords=data_array.coords, dims=dims, name=data_array.name,
                                  attrs=data_array.attrs)
        return data_array

    def apply_accumulate_index_offset(self, data_array, ds_info):
        """Apply the accumulate_index_offset transform on a given array."""
        # retrieve the __index_offset here, or create it if missing:
        # And keep track of the shared ds_info dict to reset it later in combine_info()
        self.current_ds_info = ds_info
        offset = ds_info.setdefault("__index_offset", 0)

        ref_var = self.get_transform_reference("accumulate_index_offset", ds_info)

        # Apply the current index_offset already reached on the indices we have in the current dataset:
        data_array = data_array + offset

        # Now update the __index_offset adding the number of elements in the reference array:
        ds_info["__index_offset"] = offset + ref_var.size
        logger.debug("Adding %d elements for index offset, new value is: %d",
                     ref_var.size, ds_info["__index_offset"])

        return data_array

    def apply_seconds_to_datetime(self, data_array, ds_info):
        """Apply the seconds_to_datetime transform on a given array."""
        # Retrieve the epoch timestamp:
        epoch_ts = np.datetime64("2000-01-01T00:00:00.000000")

        # And add our values as delta times in seconds:
        # note that we use a resolution of 1ns here:
        data_array = epoch_ts + (data_array * 1e9).astype("timedelta64[ns]")
        return data_array

    def apply_seconds_to_timedelta(self, data_array, _ds_info):
        """Apply the seconds_to_timedelta transform on a given array."""
        # Apply the type conversion in place in the data_array:
        # note that we use a resolution of 1ns here:
        data_array = (data_array * 1e9).astype("timedelta64[ns]")
        return data_array

    def apply_milliseconds_to_timedelta(self, data_array, _ds_info):
        """Apply the milliseconds_to_timedelta transform on a given array."""
        # Apply the type conversion in place in the data_array:
        # note that we use a resolution of 1ns here:
        data_array = (data_array * 1e6).astype("timedelta64[ns]")
        return data_array

    def get_transform_reference(self, transform_name, ds_info):
        """Retrieve a variable that should be used as reference during a transform."""
        var_path = ds_info[transform_name]

        if "{sector_name}" in var_path:
            # We really expect to have a sector name for that variable:
            var_path = var_path.replace("{sector_name}", ds_info["sector_name"])

        # get the variable on that path:
        ref_var = self.get_measured_variable(var_path)

        return ref_var

    def apply_transforms(self, data_array, ds_info):
        """Apply all transformations requested in the ds_info on the provided data array."""
        # Rescaling should be enabled by default:
        ds_info.setdefault("use_rescaling", True)
        for tname in self.transform_names:
            if tname in ds_info:
                # Retrieve the transform function:
                transform = getattr(self, f"apply_{tname}")
                # Apply the transformation on the dataset:
                data_array = transform(data_array, ds_info)
        return data_array

    def combine_info(self, all_infos):
        """Re-implement combine_info.

        This is to be able to reset our __index_offset attribute in the shared ds_info currently being updated.
        """
        if self.current_ds_info is not None:
            del self.current_ds_info["__index_offset"]
            self.current_ds_info = None

        return super().combine_info(all_infos)

    def get_transformed_dataset(self, ds_info):
        """Retrieve a dataset with all transformations applied on it."""
        # Extract base variable name:
        vname = ds_info["variable_name"]

        # Note that the sector name might be None below:
        sname = ds_info.get("sector_name", None)

        # Use the sector name as prefix for the variable path if applicable:
        var_paths = vname if sname is None else f"{sname}/{vname}"

        # Note that this includes the case where sname == None:
        data_array = self.get_measured_variable(var_paths)
        data_array = self.apply_transforms(data_array, ds_info)
        return data_array

    def validate_array_dimensions(self, data_array, ds_info=None):
        """Ensure that the dimensions of the provided data_array are valid."""
        # We also need a special handling of the ndim==0 case (i.e. reading scalar values)
        # in order to potentially support data array combination in a satpy scene:
        if data_array.ndim == 0:
            # If we have no dimension, we should force creating one here:
            data_array = data_array.expand_dims({"y": 1})

        data_array = data_array.rename({data_array.dims[0]: "y"})

        return data_array

    def update_array_attributes(self, data_array, ds_info):
        """Inject the attributes from the ds_info structure into the final data array, ignoring the internal entries."""
        # ignore some internal processing only entries:
        ignored_attribs = ["__index_offset", "broadcast_to", "accumulate_index_offset",
                           "seconds_to_timedelta", "seconds_to_datetime"]
        for key, value in ds_info.items():
            if key not in ignored_attribs:
                data_array.attrs[key] = value

        return data_array

    def get_dataset(self, dataset_id, ds_info=None):
        """Get a dataset."""
        # Retrieve default infos if missing:
        if ds_info is None:
            ds_info = self.get_dataset_infos(dataset_id["name"])
        # check for potential error:
        if ds_info is None:
            raise KeyError(f"No dataset registered for {dataset_id}")

        ds_name = ds_info["name"]
        # In case this dataset name is not explicitly provided by this file handler then we
        # should simply return None.
        if ds_name not in self.provided_datasets:
            return None

        # Generate our coordinates from azimuth/elevation data if needed.
        # It shall be called only when a corresponding dataset is being requested
        # (i.e. longitude and latitude for accumulated products)
        coord_names = self.get_latlon_names()
        is_coord = ds_name in coord_names
        # call only when internal variable is empty, to avoid multiple call.
        if ds_name not in self.internal_variables and is_coord and self.prod_in_accumulation_grid:
            self.generate_coords_from_scan_angles()

        # Retrieve the transformed data array:
        data_array = self.get_transformed_dataset(ds_info)

        # Validate the dimensions:
        data_array = self.validate_array_dimensions(data_array, ds_info)

        # Update the attributes in the final array:
        data_array = self.update_array_attributes(data_array, ds_info)

        # Return the resulting array:
        return data_array



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""CF processing of time information (coordinates and dimensions)."""
import datetime as dt
import logging

import numpy as np
import pytest
import xarray as xr
from pyresample import AreaDefinition

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - caplog


class TestCFtime:
    """Test cases for CF time dimension and coordinates."""

    @pytest.fixture
    def times(self):
        """Get time coordinate."""
        return np.array(["2018-05-30T10:00:00", "2018-05-30T10:15:00"], dtype="datetime64[ns]")

    @pytest.fixture
    def data_array_with_attrs(self, times):
        """Make fake data array with time attributes."""
        return xr.DataArray(
            np.array([[1, 2], [3, 4]]),
            dims=["time", "y"],
            coords={"time": times},
            attrs={
                "start_time": dt.datetime(2018, 5, 30, 10),
                "end_time": dt.datetime(2018,5, 30, 10, 28, 33)
            }
        )

    @pytest.fixture
    def data_array_without_attrs(self, data_array_with_attrs):
        """Make fake data array without attributes."""
        arr = data_array_with_attrs.copy()
        arr.attrs = {}
        return arr

    @pytest.fixture
    def fake_dataset(self, data_array_with_attrs, data_array_without_attrs):
        """Make fake dataset."""
        return xr.Dataset(
            {
                "with_attrs": data_array_with_attrs,
                "without_attrs": data_array_without_attrs
            }
        )

    @pytest.fixture
    def dataset_exp(self, times, data_array_with_attrs, data_array_without_attrs):
        """Make expected dataset with time bounds."""
        time_bounds_exp = xr.DataArray(
            np.array([
                ["2018-05-30T10:00:00", "2018-05-30T10:15:00"],
                ["2018-05-30T10:15:00", "2018-05-30T10:28:33"]
            ], dtype="datetime64[ns]"),
            dims=("time", "bnds_1d"),
            coords={"time": times},
            name="time_bnds"
        )
        return xr.Dataset(
            {
                "with_attrs": data_array_with_attrs,
                "without_attrs": data_array_without_attrs,
                "time_bnds": time_bounds_exp
            },
            coords={
                "time": xr.DataArray(
                    times,
                    dims="time",
                    attrs={
                        "bounds": "time_bnds",
                        "standard_name": "time"
                    }
                )
            }
        )

    def test_adding_time_bounds_to_timeseries(self, fake_dataset, dataset_exp):
        """Test adding time bounds for multiple timesteps."""
        from satpy.cf.coords import add_time_bounds_dimension
        ds_with_bounds = add_time_bounds_dimension(fake_dataset)
        xr.testing.assert_identical(ds_with_bounds, dataset_exp)

    def test_add_time_bounds_dimension(self):
        """Test addition of CF-compliant time attributes."""
        from satpy.cf.coords import add_time_bounds_dimension

        test_array = np.array([[1, 2], [3, 4], [5, 6], [7, 8]])
        times = np.array(["2018-05-30T10:05:00", "2018-05-30T10:05:01",
                          "2018-05-30T10:05:02", "2018-05-30T10:05:03"], dtype=np.datetime64)
        dataarray = xr.DataArray(test_array,
                                 dims=["y", "x"],
                                 coords={"time": ("y", times)},
                                 attrs=dict(start_time=times[0], end_time=times[-1]))
        ds = dataarray.to_dataset(name="test-array")
        ds = add_time_bounds_dimension(ds)

        assert "bnds_1d" in ds.dims
        assert ds.sizes["bnds_1d"] == 2
        assert "time_bnds" in list(ds.data_vars)
        assert "bounds" in ds["time"].attrs
        assert "standard_name" in ds["time"].attrs


class TestCFcoords:
    """Test cases for CF spatial dimension and coordinates."""

    def test_check_unique_projection_coords(self):
        """Test that the x and y coordinates are unique."""
        from satpy.cf.coords import check_unique_projection_coords

        dummy = [[1, 2], [3, 4]]
        datas = {"a": xr.DataArray(data=dummy, dims=("y", "x"), coords={"y": [1, 2], "x": [3, 4]}),
                 "b": xr.DataArray(data=dummy, dims=("y", "x"), coords={"y": [1, 2], "x": [3, 4]}),
                 "n": xr.DataArray(data=dummy, dims=("v", "w"), coords={"v": [1, 2], "w": [3, 4]})}
        check_unique_projection_coords(datas)

        datas["c"] = xr.DataArray(data=dummy, dims=("y", "x"), coords={"y": [1, 3], "x": [3, 4]})
        with pytest.raises(ValueError, match="must have identical projection coordinates"):
            check_unique_projection_coords(datas)

    def test_add_coordinates_attrs_coords(self):
        """Check that coordinates link has been established correctly."""
        from satpy.cf.coords import add_coordinates_attrs_coords

        data = [[1, 2], [3, 4]]
        lon = np.zeros((2, 2))
        lon2 = np.zeros((1, 2, 2))
        lat = np.ones((2, 2))
        datasets = {
            "var1": xr.DataArray(data=data, dims=("y", "x"), attrs={"coordinates": "lon lat"}),
            "var2": xr.DataArray(data=data, dims=("y", "x")),
            "var3": xr.DataArray(data=data, dims=("y", "x"), attrs={"coordinates": "lon2 lat"}),
            "var4": xr.DataArray(data=data, dims=("y", "x"), attrs={"coordinates": "not_exist lon lat"}),
            "lon": xr.DataArray(data=lon, dims=("y", "x")),
            "lon2": xr.DataArray(data=lon2, dims=("time", "y", "x")),
            "lat": xr.DataArray(data=lat, dims=("y", "x"))
        }

        with pytest.warns(UserWarning, match="""Coordinate \"not_exist\" referenced"""):
            datasets = add_coordinates_attrs_coords(datasets)

        # Check that link has been established correctly and 'coordinate' atrribute has been dropped
        assert "lon" in datasets["var1"].coords
        assert "lat" in datasets["var1"].coords
        np.testing.assert_array_equal(datasets["var1"]["lon"].data, lon)
        np.testing.assert_array_equal(datasets["var1"]["lat"].data, lat)
        assert "coordinates" not in datasets["var1"].attrs

        # There should be no link if there was no 'coordinate' attribute
        assert "lon" not in datasets["var2"].coords
        assert "lat" not in datasets["var2"].coords

        # The non-existent dimension or coordinate should be dropped
        assert "time" not in datasets["var3"].coords
        assert "not_exist" not in datasets["var4"].coords

    def test_ensure_unique_nondimensional_coords(self):
        """Test that created coordinate variables are unique."""
        from satpy.cf.coords import ensure_unique_nondimensional_coords

        data = [[1, 2], [3, 4]]
        y = [1, 2]
        x = [1, 2]
        time1 = [1, 2]
        time2 = [3, 4]
        datasets = {"var1": xr.DataArray(data=data,
                                         dims=("y", "x"),
                                         coords={"y": y, "x": x, "acq_time": ("y", time1)}),
                    "var2": xr.DataArray(data=data,
                                         dims=("y", "x"),
                                         coords={"y": y, "x": x, "acq_time": ("y", time2)})}

        # Test that dataset names are prepended to alternative coordinates
        res = ensure_unique_nondimensional_coords(datasets)
        np.testing.assert_array_equal(res["var1"]["var1_acq_time"], time1)
        np.testing.assert_array_equal(res["var2"]["var2_acq_time"], time2)
        assert "acq_time" not in res["var1"].coords
        assert "acq_time" not in res["var2"].coords

        # Make sure nothing else is modified
        np.testing.assert_array_equal(res["var1"]["x"], x)
        np.testing.assert_array_equal(res["var1"]["y"], y)
        np.testing.assert_array_equal(res["var2"]["x"], x)
        np.testing.assert_array_equal(res["var2"]["y"], y)

        # Coords not unique -> Dataset names must be prepended, even if pretty=True
        with pytest.warns(UserWarning, match='Cannot pretty-format "acq_time"'):
            res = ensure_unique_nondimensional_coords(datasets, pretty=True)
        np.testing.assert_array_equal(res["var1"]["var1_acq_time"], time1)
        np.testing.assert_array_equal(res["var2"]["var2_acq_time"], time2)
        assert "acq_time" not in res["var1"].coords
        assert "acq_time" not in res["var2"].coords

        # Coords unique and pretty=True -> Don't modify coordinate names
        datasets["var2"]["acq_time"] = ("y", time1)
        res = ensure_unique_nondimensional_coords(datasets, pretty=True)
        np.testing.assert_array_equal(res["var1"]["acq_time"], time1)
        np.testing.assert_array_equal(res["var2"]["acq_time"], time1)
        assert "var1_acq_time" not in res["var1"].coords
        assert "var2_acq_time" not in res["var2"].coords

    def test_is_projected(self, caplog):
        """Tests for private _is_projected function."""
        from satpy.cf.coords import _is_projected

        # test case with units but no area
        da = xr.DataArray(
            np.arange(25).reshape(5, 5),
            dims=("y", "x"),
            coords={"x": xr.DataArray(np.arange(5), dims=("x",), attrs={"units": "m"}),
                    "y": xr.DataArray(np.arange(5), dims=("y",), attrs={"units": "m"})})
        assert _is_projected(da)

        da = xr.DataArray(
            np.arange(25).reshape(5, 5),
            dims=("y", "x"),
            coords={"x": xr.DataArray(np.arange(5), dims=("x",), attrs={"units": "degrees_east"}),
                    "y": xr.DataArray(np.arange(5), dims=("y",), attrs={"units": "degrees_north"})})
        assert not _is_projected(da)

        da = xr.DataArray(
            np.arange(25).reshape(5, 5),
            dims=("y", "x"))
        with caplog.at_level(logging.WARNING):
            assert _is_projected(da)
        assert "Failed to tell if data are projected." in caplog.text

    @pytest.fixture
    def datasets(self):
        """Create test dataset."""
        data = [[75, 2], [3, 4]]
        y = [1, 2]
        x = [1, 2]
        geos = AreaDefinition(
            area_id="geos",
            description="geos",
            proj_id="geos",
            projection={"proj": "geos", "h": 35785831., "a": 6378169., "b": 6356583.8},
            width=2, height=2,
            area_extent=[-1, -1, 1, 1])
        datasets = {
            "var1": xr.DataArray(data=data,
                                 dims=("y", "x"),
                                 coords={"y": y, "x": x}),
            "var2": xr.DataArray(data=data,
                                 dims=("y", "x"),
                                 coords={"y": y, "x": x}),
            "lat": xr.DataArray(data=data,
                                dims=("y", "x"),
                                coords={"y": y, "x": x}),
            "lon": xr.DataArray(data=data,
                                dims=("y", "x"),
                                coords={"y": y, "x": x})}
        datasets["lat"].attrs["standard_name"] = "latitude"
        datasets["var1"].attrs["standard_name"] = "dummy"
        datasets["var2"].attrs["standard_name"] = "dummy"
        datasets["var2"].attrs["area"] = geos
        datasets["var1"].attrs["area"] = geos
        datasets["lat"].attrs["name"] = "lat"
        datasets["var1"].attrs["name"] = "var1"
        datasets["var2"].attrs["name"] = "var2"
        datasets["lon"].attrs["name"] = "lon"
        return datasets

    def test__is_lon_or_lat_dataarray(self, datasets):
        """Test the _is_lon_or_lat_dataarray function."""
        from satpy.cf.coords import _is_lon_or_lat_dataarray

        assert _is_lon_or_lat_dataarray(datasets["lat"])
        assert not _is_lon_or_lat_dataarray(datasets["var1"])

    def test_has_projection_coords(self, datasets):
        """Test the has_projection_coords function."""
        from satpy.cf.coords import has_projection_coords

        assert has_projection_coords(datasets)
        datasets["lat"].attrs["standard_name"] = "dummy"
        assert not has_projection_coords(datasets)



#!/usr/bin/env python
# Copyright (c) 2018-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for arithmetic compositors."""

import datetime as dt
import unittest

import dask.array as da
import numpy as np
import pytest
import xarray as xr


class TestDifferenceCompositor(unittest.TestCase):
    """Test case for the difference compositor."""

    def setUp(self):
        """Create test data."""
        from pyresample.geometry import AreaDefinition
        area = AreaDefinition("test", "test", "test",
                              {"proj": "merc"}, 2, 2,
                              (-2000, -2000, 2000, 2000))
        attrs = {"area": area,
                 "start_time": dt.datetime(2018, 1, 1, 18),
                 "modifiers": tuple(),
                 "resolution": 1000,
                 "name": "test_vis"}
        ds1 = xr.DataArray(da.ones((2, 2), chunks=2, dtype=np.float64),
                           attrs=attrs, dims=("y", "x"),
                           coords={"y": [0, 1], "x": [0, 1]})
        self.ds1 = ds1
        ds2 = xr.DataArray(da.ones((2, 2), chunks=2, dtype=np.float64) + 2,
                           attrs=attrs, dims=("y", "x"),
                           coords={"y": [0, 1], "x": [0, 1]})
        ds2.attrs["name"] += "2"
        self.ds2 = ds2

        # high res version
        ds2 = xr.DataArray(da.ones((4, 4), chunks=2, dtype=np.float64) + 4,
                           attrs=attrs.copy(), dims=("y", "x"),
                           coords={"y": [0, 1, 2, 3], "x": [0, 1, 2, 3]})
        ds2.attrs["name"] += "2"
        ds2.attrs["resolution"] = 500
        ds2.attrs["rows_per_scan"] = 1
        ds2.attrs["area"] = AreaDefinition("test", "test", "test",
                                           {"proj": "merc"}, 4, 4,
                                           (-2000, -2000, 2000, 2000))
        self.ds2_big = ds2

    def test_basic_diff(self):
        """Test that a basic difference composite works."""
        from satpy.composites.arithmetic import DifferenceCompositor
        comp = DifferenceCompositor(name="diff", standard_name="temperature_difference")
        res = comp((self.ds1, self.ds2))
        np.testing.assert_allclose(res.values, -2)
        assert res.attrs.get("standard_name") == "temperature_difference"

    def test_bad_areas_diff(self):
        """Test that a difference where resolutions are different fails."""
        from satpy.composites.arithmetic import DifferenceCompositor
        from satpy.composites.core import IncompatibleAreas
        comp = DifferenceCompositor(name="diff")
        # too many arguments
        with pytest.raises(ValueError, match="Expected 2 datasets, got 3"):
            comp((self.ds1, self.ds2, self.ds2_big))
        # different resolution
        with pytest.raises(IncompatibleAreas):
            comp((self.ds1, self.ds2_big))


@pytest.fixture
def fake_area():
    """Return a fake 2×2 area."""
    from pyresample.geometry import create_area_def
    return create_area_def("skierffe", 4087, area_extent=[-5_000, -5_000, 5_000, 5_000], shape=(2, 2))


@pytest.fixture
def fake_dataset_pair(fake_area):
    """Return a fake pair of 2×2 datasets."""
    ds1 = xr.DataArray(da.full((2, 2), 8, chunks=2, dtype=np.float32),
                       attrs={"area": fake_area, "standard_name": "toa_bidirectional_reflectance"})
    ds2 = xr.DataArray(da.full((2, 2), 4, chunks=2, dtype=np.float32),
                       attrs={"area": fake_area, "standard_name": "toa_bidirectional_reflectance"})
    return (ds1, ds2)


@pytest.mark.parametrize("kwargs", [{}, {"standard_name": "channel_ratio", "foo": "bar"}])
def test_ratio_compositor(fake_dataset_pair, kwargs):
    """Test the ratio compositor."""
    from satpy.composites.arithmetic import RatioCompositor
    comp = RatioCompositor("ratio", **kwargs)
    res = comp(fake_dataset_pair)
    np.testing.assert_allclose(res.values, 2)

    assert res.attrs["name"] == "ratio"

    if "standard_name" in kwargs:
        # See that the kwargs have been updated to the attrs
        assert res.attrs["standard_name"] == "channel_ratio"
        assert res.attrs["foo"] == "bar"
    else:
        assert res.attrs["standard_name"] == "toa_bidirectional_reflectance"


def test_sum_compositor(fake_dataset_pair):
    """Test the sum compositor."""
    from satpy.composites.arithmetic import SumCompositor
    comp = SumCompositor(name="sum", standard_name="channel_sum")
    res = comp(fake_dataset_pair)
    np.testing.assert_allclose(res.values, 12)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Benchmark VIIRS SDR operations.."""

from __future__ import annotations

import glob
import os

from pyspectral.rayleigh import check_and_download as download_luts
from pyspectral.rsr_reader import check_and_download as download_rsr


class VIIRSSDRBenchmarkBase:
    """Shared methods for working with VIIRS SDR data."""

    timeout = 600
    data_files: list[str] = []

    def setup_cache(self):
        """Fetch the data files."""
        try:
            from satpy.demo import get_viirs_sdr_20170128_1229
            get_viirs_sdr_20170128_1229(
                channels=("I01", "M03", "M04", "M05"),
                granules=(2, 3, 4))
        except ImportError:
            assert len(self.get_filenames()) == 6 * 3  # nosec
        download_rsr()
        download_luts(aerosol_types=["rayleigh_only"])

    def setup(self, name):
        """Set up the benchmarks."""
        import dask.config
        self.data_files = self.get_filenames()
        dask.config.set({"array.chunk-size": "32MiB"})

    def get_filenames(self):
        """Get the data filenames manually."""
        base_dir = os.environ.get("SATPY_DEMO_DATA_DIR", ".")
        return glob.glob(os.path.join(base_dir, "viirs_sdr", "20170128_1229", "*.h5"))

    def load(self, composite):
        """Load one composite."""
        from satpy import Scene
        scn = Scene(filenames=self.data_files, reader="viirs_sdr")
        scn.load([composite])
        return scn

    def load_and_native_resample(self, composite):
        """Load and native resample a composite."""
        scn = self.load(composite)
        lscn = scn.resample(resampler="native")
        return lscn


class VIIRSSDRReaderBenchmarks(VIIRSSDRBenchmarkBase):
    """Benchmark reading and writing VIIRS SDR data."""

    params = ["I01", "M03"]
    param_names = ["name"]

    def time_load_one_channel(self, name):
        """Time the loading of one channel."""
        self.compute_product(name)

    def peakmem_load_one_channel(self, name):
        """Check peak memory usage of loading one channel."""
        self.compute_product(name)

    def compute_product(self, name):
        """Load and compute one channel."""
        scn = self.load(name)
        scn[name].compute()


class VIIRSSDRCompositeBenchmarks(VIIRSSDRBenchmarkBase):
    """Benchmark generating and writing composites from VIIRS SDR data."""

    params = ["true_color", "true_color_crefl", "true_color_raw"]
    param_names = ["name"]

    def time_load_composite(self, name):
        """Time the loading of the generation of a composite."""
        self.compute_composite(name)

    def peakmem_load_composite(self, name):
        """Check peak memory usage of the generation of a composite."""
        self.compute_composite(name)

    def time_save_composite_to_geotiff(self, name):
        """Time the generation and saving of a composite."""
        self.save_composite_as_geotiff(name)

    def peakmem_save_composite_raw_to_geotiff(self, name):
        """Check peak memory usage of the generation and saving of a composite."""
        self.save_composite_as_geotiff(name)

    def compute_composite(self, name):
        """Compute a composite."""
        lscn = self.load_and_native_resample(name)
        lscn[name].compute()

    def save_composite_as_geotiff(self, name):
        """Save a composite to disk as geotiff."""
        lscn = self.load_and_native_resample(name)
        lscn.save_dataset(name, filename="test.tif", tiled=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Interface to VIIRS flood product."""

import numpy as np
from pyresample import geometry

from satpy.readers.core.hdf4 import HDF4FileHandler


class VIIRSEDRFlood(HDF4FileHandler):
    """VIIRS EDR Flood-product handler for HDF4 files."""

    @property
    def start_time(self):
        """Get start time."""
        return self.filename_info["start_time"]

    @property
    def end_time(self):
        """Get end time."""
        return self.filename_info.get("end_time", self.start_time)

    @property
    def sensor_name(self):
        """Get sensor name."""
        sensor = self["/attr/SensorIdentifyCode"]
        if isinstance(sensor, np.ndarray):
            return str(sensor.astype(str)).lower()
        return sensor.lower()

    @property
    def platform_name(self):
        """Get platform name."""
        platform_name = self["/attr/Satellitename"]
        if isinstance(platform_name, np.ndarray):
            return str(platform_name.astype(str)).lower()
        return platform_name.lower()

    def get_metadata(self, data, ds_info):
        """Get metadata."""
        metadata = {}
        metadata.update(data.attrs)
        metadata.update(ds_info)
        metadata.update({
            "sensor": self.sensor_name,
            "platform_name": self.platform_name,
            "start_time": self.start_time,
            "end_time": self.end_time,
        })

        return metadata

    def get_dataset(self, ds_id, ds_info):
        """Get dataset."""
        data = self[ds_id["name"]]

        data.attrs = self.get_metadata(data, ds_info)

        fill = data.attrs.pop("_Fillvalue")
        offset = data.attrs.get("add_offset")
        scale_factor = data.attrs.get("scale_factor")

        data = data.where(data != fill)
        if scale_factor is not None and offset is not None:
            data *= scale_factor
            data += offset

        return data

    def get_area_def(self, ds_id):
        """Get area definition."""
        data = self[ds_id["name"]]

        proj_dict = {
            "proj": "latlong",
            "datum": "WGS84",
            "ellps": "WGS84",
            "no_defs": True
        }

        area_extent = [data.attrs.get("ProjectionMinLongitude"), data.attrs.get("ProjectionMinLatitude"),
                       data.attrs.get("ProjectionMaxLongitude"), data.attrs.get("ProjectionMaxLatitude")]

        area = geometry.AreaDefinition(
            "viirs_flood_area",
            "name_of_proj",
            "id_of_proj",
            proj_dict,
            int(self.filename_info["dim0"]),
            int(self.filename_info["dim1"]),
            np.asarray(area_extent)
        )

        return area
